# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/dataobs/validation__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.dataobs import expectation__client_pb2 as tecton__proto_dot_dataobs_dot_expectation____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-tecton_proto/dataobs/validation__client.proto\x12\x14tecton_proto.dataobs\x1a\x1fgoogle/protobuf/timestamp.proto\x1a$tecton_proto/common/id__client.proto\x1a.tecton_proto/dataobs/expectation__client.proto\"\xd0\x06\n\x11\x45xpectationResult\x12\x43\n\x11validation_job_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fvalidationJobId\x12\x1c\n\tworkspace\x18\x02 \x01(\tR\tworkspace\x12*\n\x11\x66\x65\x61ture_view_name\x18\t \x01(\tR\x0f\x66\x65\x61tureViewName\x12\x45\n\x12\x66\x65\x61ture_package_id\x18\x03 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61turePackageId\x12\x43\n\x0fvalidation_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0evalidationTime\x12Y\n\x1b\x66\x65\x61ture_interval_start_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x18\x66\x65\x61tureIntervalStartTime\x12U\n\x19\x66\x65\x61ture_interval_end_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x16\x66\x65\x61tureIntervalEndTime\x12t\n\x1c\x66\x65\x61ture_expectation_metadata\x18\x06 \x01(\x0b\x32\x30.tecton_proto.dataobs.FeatureExpectationMetadataH\x00R\x1a\x66\x65\x61tureExpectationMetadata\x12q\n\x1bmetric_expectation_metadata\x18\x07 \x01(\x0b\x32/.tecton_proto.dataobs.MetricExpectationMetadataH\x00R\x19metricExpectationMetadata\x12\x43\n\x06result\x18\x08 \x01(\x0e\x32+.tecton_proto.dataobs.ExpectationResultEnumR\x06result\x12\x34\n\tresult_id\x18\n \x01(\x0b\x32\x17.tecton_proto.common.IdR\x08resultIdB\n\n\x08metadata\"\xeb\x01\n\x1a\x46\x65\x61tureExpectationMetadata\x12J\n\x0b\x65xpectation\x18\x01 \x01(\x0b\x32(.tecton_proto.dataobs.FeatureExpectationR\x0b\x65xpectation\x12\x1b\n\talert_msg\x18\x03 \x01(\tR\x08\x61lertMsg\x12-\n\x12\x66\x61ilure_percentage\x18\x04 \x01(\x01R\x11\x66\x61ilurePercentage\x12\x35\n\x17\x66\x61iled_join_key_samples\x18\x05 \x03(\tR\x14\x66\x61iledJoinKeySamples\"\x81\x03\n\x19MetricExpectationMetadata\x12I\n\x0b\x65xpectation\x18\x01 \x01(\x0b\x32\'.tecton_proto.dataobs.MetricExpectationR\x0b\x65xpectation\x12\x1b\n\talert_msg\x18\x03 \x01(\tR\x08\x61lertMsg\x12]\n\x0cparam_values\x18\x04 \x03(\x0b\x32:.tecton_proto.dataobs.MetricExpectationMetadata.ParamValueR\x0bparamValues\x1a\x9c\x01\n\nParamValue\x12\x1f\n\x0bmetric_name\x18\x01 \x01(\tR\nmetricName\x12!\n\x0c\x61\x63tual_value\x18\x02 \x01(\tR\x0b\x61\x63tualValue\x12J\n\x13interval_start_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11intervalStartTime\"o\n\rResultSummary\x12\x16\n\x06passed\x18\x01 \x01(\x05R\x06passed\x12\x16\n\x06\x66\x61iled\x18\x02 \x01(\x05R\x06\x66\x61iled\x12\x14\n\x05\x65rror\x18\x03 \x01(\x05R\x05\x65rror\x12\x18\n\x07unknown\x18\x04 \x01(\x05R\x07unknown\"\xd7\x01\n\x16WorkspaceResultSummary\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12=\n\x07summary\x18\x02 \x01(\x0b\x32#.tecton_proto.dataobs.ResultSummaryR\x07summary\x12`\n\x14\x66\x65\x61ture_view_summary\x18\x03 \x03(\x0b\x32..tecton_proto.dataobs.FeatureViewResultSummaryR\x12\x66\x65\x61tureViewSummary\"\xe6\x01\n\x18\x46\x65\x61tureViewResultSummary\x12*\n\x11\x66\x65\x61ture_view_name\x18\x01 \x01(\tR\x0f\x66\x65\x61tureViewName\x12=\n\x07summary\x18\x02 \x01(\x0b\x32#.tecton_proto.dataobs.ResultSummaryR\x07summary\x12_\n\x13\x65xpectation_summary\x18\x03 \x03(\x0b\x32..tecton_proto.dataobs.ExpectationResultSummaryR\x12\x65xpectationSummary\"\x84\x01\n\x18\x45xpectationResultSummary\x12)\n\x10\x65xpectation_name\x18\x01 \x01(\tR\x0f\x65xpectationName\x12=\n\x07summary\x18\x02 \x01(\x0b\x32#.tecton_proto.dataobs.ResultSummaryR\x07summary*c\n\x15\x45xpectationResultEnum\x12\x12\n\x0eRESULT_UNKNOWN\x10\x00\x12\x11\n\rRESULT_PASSED\x10\x01\x12\x11\n\rRESULT_FAILED\x10\x02\x12\x10\n\x0cRESULT_ERROR\x10\x03\x42\x16\n\x12\x63om.tecton.dataobsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.dataobs.validation__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.tecton.dataobsP\001'
  _EXPECTATIONRESULTENUM._serialized_start=2366
  _EXPECTATIONRESULTENUM._serialized_end=2465
  _EXPECTATIONRESULT._serialized_start=191
  _EXPECTATIONRESULT._serialized_end=1039
  _FEATUREEXPECTATIONMETADATA._serialized_start=1042
  _FEATUREEXPECTATIONMETADATA._serialized_end=1277
  _METRICEXPECTATIONMETADATA._serialized_start=1280
  _METRICEXPECTATIONMETADATA._serialized_end=1665
  _METRICEXPECTATIONMETADATA_PARAMVALUE._serialized_start=1509
  _METRICEXPECTATIONMETADATA_PARAMVALUE._serialized_end=1665
  _RESULTSUMMARY._serialized_start=1667
  _RESULTSUMMARY._serialized_end=1778
  _WORKSPACERESULTSUMMARY._serialized_start=1781
  _WORKSPACERESULTSUMMARY._serialized_end=1996
  _FEATUREVIEWRESULTSUMMARY._serialized_start=1999
  _FEATUREVIEWRESULTSUMMARY._serialized_end=2229
  _EXPECTATIONRESULTSUMMARY._serialized_start=2232
  _EXPECTATIONRESULTSUMMARY._serialized_end=2364
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/secrets/secrets_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auditlog import metadata__client_pb2 as tecton__proto_dot_auditlog_dot_metadata____client__pb2
from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/secrets/secrets_service__client.proto\x12\x14tecton_proto.secrets\x1a,tecton_proto/auditlog/metadata__client.proto\x1a\'tecton_proto/auth/service__client.proto\"!\n\x0bSecretScope\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"\x1f\n\tSecretKey\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"\x19\n\x17ListSecretScopesRequest\"U\n\x18ListSecretScopesResponse\x12\x39\n\x06scopes\x18\x01 \x03(\x0b\x32!.tecton_proto.secrets.SecretScopeR\x06scopes\"9\n\x18\x43reateSecretScopeRequest\x12\x1d\n\x05scope\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x05scope\"\x1b\n\x19\x43reateSecretScopeResponse\"9\n\x18\x44\x65leteSecretScopeRequest\x12\x1d\n\x05scope\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x05scope\"\x1b\n\x19\x44\x65leteSecretScopeResponse\"*\n\x12ListSecretsRequest\x12\x14\n\x05scope\x18\x01 \x01(\tR\x05scope\"J\n\x13ListSecretsResponse\x12\x33\n\x04keys\x18\x02 \x03(\x0b\x32\x1f.tecton_proto.secrets.SecretKeyR\x04keys\"?\n\x15GetSecretValueRequest\x12\x14\n\x05scope\x18\x01 \x01(\tR\x05scope\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\".\n\x16GetSecretValueResponse\x12\x14\n\x05value\x18\x01 \x01(\tR\x05value\"w\n\x15PutSecretValueRequest\x12\x1d\n\x05scope\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x05scope\x12\x19\n\x03key\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x03key\x12$\n\x05value\x18\x03 \x01(\tB\x0e\xea\x99\xbd\x46\x02\x08\x02\xea\x99\xbd\x46\x02\x10\x02R\x05value\"\x18\n\x16PutSecretValueResponse\"O\n\x13\x44\x65leteSecretRequest\x12\x1d\n\x05scope\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x05scope\x12\x19\n\x03key\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x03key\"\x16\n\x14\x44\x65leteSecretResponse2\xee\t\n\x0eSecretsService\x12{\n\x10ListSecretScopes\x12-.tecton_proto.secrets.ListSecretScopesRequest\x1a..tecton_proto.secrets.ListSecretScopesResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xc4\x01\n\x11\x43reateSecretScope\x12..tecton_proto.secrets.CreateSecretScopeRequest\x1a/.tecton_proto.secrets.CreateSecretScopeResponse\"N\xa2\xbc\xe6\xc0\x05)*\'tecton/organization#create_secret_scope\xf2\x99\xbd\x46\x1a\x08\x01\x12\x13\x63reate_secret_scope\x1a\x01\x31\x12\xcf\x01\n\x11\x44\x65leteSecretScope\x12..tecton_proto.secrets.DeleteSecretScopeRequest\x1a/.tecton_proto.secrets.DeleteSecretScopeResponse\"Y\xa2\xbc\xe6\xc0\x05\x34\x1a\t\x08\x04\x12\x05scope*\'tecton/secret_scope#delete_secret_scope\xf2\x99\xbd\x46\x1a\x08\x01\x12\x13\x64\x65lete_secret_scope\x1a\x01\x31\x12\x9a\x01\n\x0bListSecrets\x12(.tecton_proto.secrets.ListSecretsRequest\x1a).tecton_proto.secrets.ListSecretsResponse\"6\xa2\xbc\xe6\xc0\x05\x30\x1a\t\x08\x04\x12\x05scope*#tecton/secret_scope#list_secret_key\x12\xa5\x01\n\x0eGetSecretValue\x12+.tecton_proto.secrets.GetSecretValueRequest\x1a,.tecton_proto.secrets.GetSecretValueResponse\"8\xa2\xbc\xe6\xc0\x05\x32\x1a\t\x08\x04\x12\x05scope*%tecton/secret_scope#read_secret_value\x12\xc3\x01\n\x0ePutSecretValue\x12+.tecton_proto.secrets.PutSecretValueRequest\x1a,.tecton_proto.secrets.PutSecretValueResponse\"V\xa2\xbc\xe6\xc0\x05\x34\x1a\t\x08\x04\x12\x05scope*\'tecton/secret_scope#manage_secret_value\xf2\x99\xbd\x46\x17\x08\x01\x12\x10put_secret_value\x1a\x01\x31\x12\xba\x01\n\x0c\x44\x65leteSecret\x12).tecton_proto.secrets.DeleteSecretRequest\x1a*.tecton_proto.secrets.DeleteSecretResponse\"S\xa2\xbc\xe6\xc0\x05\x34\x1a\t\x08\x04\x12\x05scope*\'tecton/secret_scope#manage_secret_value\xf2\x99\xbd\x46\x14\x08\x01\x12\rdelete_secret\x1a\x01\x31\x42V\n\x12\x63om.tecton.secretsB\x13SecretsServiceProtoP\x01Z)github.com/tecton-ai/tecton_proto/secrets')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.secrets.secrets_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.tecton.secretsB\023SecretsServiceProtoP\001Z)github.com/tecton-ai/tecton_proto/secrets'
  _CREATESECRETSCOPEREQUEST.fields_by_name['scope']._options = None
  _CREATESECRETSCOPEREQUEST.fields_by_name['scope']._serialized_options = b'\352\231\275F\002\010\002'
  _DELETESECRETSCOPEREQUEST.fields_by_name['scope']._options = None
  _DELETESECRETSCOPEREQUEST.fields_by_name['scope']._serialized_options = b'\352\231\275F\002\010\002'
  _PUTSECRETVALUEREQUEST.fields_by_name['scope']._options = None
  _PUTSECRETVALUEREQUEST.fields_by_name['scope']._serialized_options = b'\352\231\275F\002\010\002'
  _PUTSECRETVALUEREQUEST.fields_by_name['key']._options = None
  _PUTSECRETVALUEREQUEST.fields_by_name['key']._serialized_options = b'\352\231\275F\002\010\002'
  _PUTSECRETVALUEREQUEST.fields_by_name['value']._options = None
  _PUTSECRETVALUEREQUEST.fields_by_name['value']._serialized_options = b'\352\231\275F\002\010\002\352\231\275F\002\020\002'
  _DELETESECRETREQUEST.fields_by_name['scope']._options = None
  _DELETESECRETREQUEST.fields_by_name['scope']._serialized_options = b'\352\231\275F\002\010\002'
  _DELETESECRETREQUEST.fields_by_name['key']._options = None
  _DELETESECRETREQUEST.fields_by_name['key']._serialized_options = b'\352\231\275F\002\010\002'
  _SECRETSSERVICE.methods_by_name['ListSecretScopes']._options = None
  _SECRETSSERVICE.methods_by_name['ListSecretScopes']._serialized_options = b'\242\274\346\300\005\002@\001'
  _SECRETSSERVICE.methods_by_name['CreateSecretScope']._options = None
  _SECRETSSERVICE.methods_by_name['CreateSecretScope']._serialized_options = b'\242\274\346\300\005)*\'tecton/organization#create_secret_scope\362\231\275F\032\010\001\022\023create_secret_scope\032\0011'
  _SECRETSSERVICE.methods_by_name['DeleteSecretScope']._options = None
  _SECRETSSERVICE.methods_by_name['DeleteSecretScope']._serialized_options = b'\242\274\346\300\0054\032\t\010\004\022\005scope*\'tecton/secret_scope#delete_secret_scope\362\231\275F\032\010\001\022\023delete_secret_scope\032\0011'
  _SECRETSSERVICE.methods_by_name['ListSecrets']._options = None
  _SECRETSSERVICE.methods_by_name['ListSecrets']._serialized_options = b'\242\274\346\300\0050\032\t\010\004\022\005scope*#tecton/secret_scope#list_secret_key'
  _SECRETSSERVICE.methods_by_name['GetSecretValue']._options = None
  _SECRETSSERVICE.methods_by_name['GetSecretValue']._serialized_options = b'\242\274\346\300\0052\032\t\010\004\022\005scope*%tecton/secret_scope#read_secret_value'
  _SECRETSSERVICE.methods_by_name['PutSecretValue']._options = None
  _SECRETSSERVICE.methods_by_name['PutSecretValue']._serialized_options = b'\242\274\346\300\0054\032\t\010\004\022\005scope*\'tecton/secret_scope#manage_secret_value\362\231\275F\027\010\001\022\020put_secret_value\032\0011'
  _SECRETSSERVICE.methods_by_name['DeleteSecret']._options = None
  _SECRETSSERVICE.methods_by_name['DeleteSecret']._serialized_options = b'\242\274\346\300\0054\032\t\010\004\022\005scope*\'tecton/secret_scope#manage_secret_value\362\231\275F\024\010\001\022\rdelete_secret\032\0011'
  _SECRETSCOPE._serialized_start=163
  _SECRETSCOPE._serialized_end=196
  _SECRETKEY._serialized_start=198
  _SECRETKEY._serialized_end=229
  _LISTSECRETSCOPESREQUEST._serialized_start=231
  _LISTSECRETSCOPESREQUEST._serialized_end=256
  _LISTSECRETSCOPESRESPONSE._serialized_start=258
  _LISTSECRETSCOPESRESPONSE._serialized_end=343
  _CREATESECRETSCOPEREQUEST._serialized_start=345
  _CREATESECRETSCOPEREQUEST._serialized_end=402
  _CREATESECRETSCOPERESPONSE._serialized_start=404
  _CREATESECRETSCOPERESPONSE._serialized_end=431
  _DELETESECRETSCOPEREQUEST._serialized_start=433
  _DELETESECRETSCOPEREQUEST._serialized_end=490
  _DELETESECRETSCOPERESPONSE._serialized_start=492
  _DELETESECRETSCOPERESPONSE._serialized_end=519
  _LISTSECRETSREQUEST._serialized_start=521
  _LISTSECRETSREQUEST._serialized_end=563
  _LISTSECRETSRESPONSE._serialized_start=565
  _LISTSECRETSRESPONSE._serialized_end=639
  _GETSECRETVALUEREQUEST._serialized_start=641
  _GETSECRETVALUEREQUEST._serialized_end=704
  _GETSECRETVALUERESPONSE._serialized_start=706
  _GETSECRETVALUERESPONSE._serialized_end=752
  _PUTSECRETVALUEREQUEST._serialized_start=754
  _PUTSECRETVALUEREQUEST._serialized_end=873
  _PUTSECRETVALUERESPONSE._serialized_start=875
  _PUTSECRETVALUERESPONSE._serialized_end=899
  _DELETESECRETREQUEST._serialized_start=901
  _DELETESECRETREQUEST._serialized_end=980
  _DELETESECRETRESPONSE._serialized_start=982
  _DELETESECRETRESPONSE._serialized_end=1004
  _SECRETSSERVICE._serialized_start=1007
  _SECRETSSERVICE._serialized_end=2269
# @@protoc_insertion_point(module_scope)



from typing import Any
from typing import Optional

import pypika
from pypika import queries
from pypika import terms
from pypika import utils

from tecton_core.query import sql_compat


# Custom functions used in Snowflake queries
ArraySlice = pypika.CustomFunction("ARRAY_SLICE", ["array", "from", "to"])
ArraySize = pypika.CustomFunction("ARRAY_SIZE", ["array"])
ArrayAgg = pypika.CustomFunction("ARRAY_AGG", ["value"])
ZeroIfNull = pypika.CustomFunction("ZEROIFNULL", ["value"])


def ArrayAggWithinGroup(col, key):
    return sql_compat.CustomQuery(f"ARRAYAGG({col}) WITHIN GROUP (ORDER BY {key})")


class SnowflakeQuery(queries.Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> "SnowflakeQuery":
        return SnowflakeQueryBuilder(**kwargs)


class SnowflakeQueryBuilder(sql_compat.CaseInsensitiveSnowflakeQueryBuilder):
    """
    Subclass of pypika SnowflakeQueryBuilder to implement Snowflake specific syntax
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._lateral = None

    @utils.builder
    def lateral(self, criterion: terms.Criterion) -> "queries.QueryBuilder":
        self._lateral = criterion

    def _lateral_sql(self, quote_char: Optional[str] = None, **kwargs: Any) -> str:
        """
        https://docs.snowflake.com/en/sql-reference/constructs/join-lateral
        In a FROM clause, the LATERAL keyword allows an inline view to reference columns from a table expression that precedes that inline view.

        Syntax:
        SELECT ...
        FROM <left_hand_table_expression>, LATERAL ( <inline_view> )
        ...
        """
        return " ,LATERAL {lateral}".format(
            lateral=self._lateral.get_sql(quote_char=quote_char, subquery=True, **kwargs)
        )

    def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs: Any) -> str:
        """
        This is a full copy of the get_sql method from pypika.queries.QueryBuilder except adding _lateral to the querystring
        pypika 0.48.9
        https://github.com/kayak/pypika/blob/master/pypika/queries.py#L1218
        """
        self._set_kwargs_defaults(kwargs)
        if not (self._selects or self._insert_table or self._delete_from or self._update_table):
            return ""
        if self._insert_table and not (self._selects or self._values):
            return ""
        if self._update_table and not self._updates:
            return ""

        has_joins = bool(self._joins)
        has_multiple_from_clauses = 1 < len(self._from)
        has_subquery_from_clause = 0 < len(self._from) and isinstance(self._from[0], queries.QueryBuilder)
        has_reference_to_foreign_table = self._foreign_table
        has_update_from = self._update_table and self._from

        kwargs["with_namespace"] = any(
            [
                has_joins,
                has_multiple_from_clauses,
                has_subquery_from_clause,
                has_reference_to_foreign_table,
                has_update_from,
            ]
        )

        if self._update_table:
            if self._with:
                querystring = self._with_sql(**kwargs)
            else:
                querystring = ""

            querystring += self._update_sql(**kwargs)

            if self._joins:
                querystring += " " + " ".join(join.get_sql(**kwargs) for join in self._joins)

            querystring += self._set_sql(**kwargs)

            if self._from:
                querystring += self._from_sql(**kwargs)

            if self._wheres:
                querystring += self._where_sql(**kwargs)

            if self._limit is not None:
                querystring += self._limit_sql()

            return querystring

        if self._delete_from:
            querystring = self._delete_sql(**kwargs)

        elif not self._select_into and self._insert_table:
            if self._with:
                querystring = self._with_sql(**kwargs)
            else:
                querystring = ""

            if self._replace:
                querystring += self._replace_sql(**kwargs)
            else:
                querystring += self._insert_sql(**kwargs)

            if self._columns:
                querystring += self._columns_sql(**kwargs)

            if self._values:
                querystring += self._values_sql(**kwargs)
                return querystring
            else:
                querystring += " " + self._select_sql(**kwargs)

        else:
            if self._with:
                querystring = self._with_sql(**kwargs)
            else:
                querystring = ""

            querystring += self._select_sql(**kwargs)

            if self._insert_table:
                querystring += self._into_sql(**kwargs)

        if self._from:
            querystring += self._from_sql(**kwargs)

        if self._using:
            querystring += self._using_sql(**kwargs)

        if self._force_indexes:
            querystring += self._force_index_sql(**kwargs)

        if self._use_indexes:
            querystring += self._use_index_sql(**kwargs)

        if self._joins:
            querystring += " " + " ".join(join.get_sql(**kwargs) for join in self._joins)

        if self._prewheres:
            querystring += self._prewhere_sql(**kwargs)

        if self._lateral:
            querystring += self._lateral_sql(**kwargs)

        if self._wheres:
            querystring += self._where_sql(**kwargs)

        if self._groupbys:
            querystring += self._group_sql(**kwargs)
            if self._mysql_rollup:
                querystring += self._rollup_sql()

        if self._havings:
            querystring += self._having_sql(**kwargs)

        if self._orderbys:
            querystring += self._orderby_sql(**kwargs)

        querystring = self._apply_pagination(querystring)

        if self._for_update:
            querystring += self._for_update_sql(**kwargs)

        if subquery:
            querystring = "({query})".format(query=querystring)

        if with_alias:
            kwargs["alias_quote_char"] = (
                self.ALIAS_QUOTE_CHAR if self.QUERY_ALIAS_QUOTE_CHAR is None else self.QUERY_ALIAS_QUOTE_CHAR
            )
            return utils.format_alias_sql(querystring, self.alias, **kwargs)

        return querystring



import datetime
import enum
from typing import Tuple

import attrs
from google.protobuf.duration_pb2 import Duration
from typeguard import typechecked

from tecton_core.errors import TectonValidationError
from tecton_core.time_utils import timedelta_to_proto
from tecton_core.time_utils import to_human_readable_str
from tecton_proto.args.feature_view__client_pb2 import TimeWindow as TimeWindowArgs
from tecton_proto.args.feature_view__client_pb2 import TimeWindowSeries as TimeWindowSeriesArgs
from tecton_proto.common.time_window__client_pb2 import LifetimeWindow
from tecton_proto.common.time_window__client_pb2 import RelativeTimeWindow
from tecton_proto.common.time_window__client_pb2 import TimeWindow
from tecton_proto.common.time_window__client_pb2 import TimeWindowSeries


__all__ = [
    "LifetimeWindowSpec",
    "RelativeTimeWindowSpec",
    "TimeWindowSpec",
    "create_time_window_spec_from_data_proto",
    "window_spec_to_window_data_proto",
]


class _WindowSortEnum(enum.IntEnum):
    TIME_WINDOW = 1
    LIFETIME = 2
    TIME_WINDOW_SERIES = 3


_TimeWindowSortKey = Tuple[_WindowSortEnum, tuple]


class TimeWindowSpec:
    def window_duration_string(self) -> str:
        """Window duration for use in aggregation name construction."""
        raise NotImplementedError

    def to_string(self) -> str:
        """Full name specification for use in secondary key aggregation output column."""
        raise NotImplementedError

    def to_sort_tuple(self) -> _TimeWindowSortKey:
        """Tuple for sorting time windows.

        This should start with the appropriate _WindowSortEnum, and then use any attributes for the specific window type for sorting.
        This approach ensures that we have a multi-layer sort where each sub-type is collocated together.
        """
        raise NotImplementedError


@attrs.frozen
class LifetimeWindowSpec(TimeWindowSpec):
    def to_proto(self) -> LifetimeWindow:
        return LifetimeWindow()

    def to_string(self) -> str:
        return "lifetime"

    def offset_string(self) -> str:
        return ""

    def window_duration_string(self) -> str:
        return "lifetime"

    def to_sort_tuple(self) -> _TimeWindowSortKey:
        return (_WindowSortEnum.LIFETIME, ())


@attrs.frozen(order=True)
class RelativeTimeWindowSpec(TimeWindowSpec):
    # window_end represents the offset (negative or zero)
    # window_start represents the offset - window_duration (negative)
    window_start: datetime.timedelta
    window_end: datetime.timedelta

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: RelativeTimeWindow) -> "RelativeTimeWindowSpec":
        return cls(
            window_start=proto.window_start.ToTimedelta(),
            window_end=proto.window_end.ToTimedelta(),
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: TimeWindowArgs) -> "RelativeTimeWindowSpec":
        return cls(
            window_start=Duration(
                seconds=proto.offset.seconds - proto.window_duration.seconds,
                nanos=proto.offset.nanos - proto.window_duration.nanos,
            ).ToTimedelta(),
            window_end=Duration(seconds=proto.offset.seconds, nanos=proto.offset.nanos).ToTimedelta(),
        )

    def to_data_proto(self) -> RelativeTimeWindow:
        return RelativeTimeWindow(
            window_start=timedelta_to_proto(self.window_start),
            window_end=timedelta_to_proto(self.window_end),
        )

    @property
    def window_duration(self) -> datetime.timedelta:
        return self.window_end - self.window_start

    @property
    def offset(self) -> datetime.timedelta:
        return self.window_end

    def to_args_proto(self) -> TimeWindowArgs:
        return TimeWindowArgs(
            window_duration=timedelta_to_proto(self.window_duration),
            offset=timedelta_to_proto(self.offset),
        )

    def to_string(self) -> str:
        offset_name = self.offset_string()
        if offset_name:
            offset_name = f"_{self.offset_string()}"
        return f"{self.window_duration_string()}{offset_name}"

    def to_sort_tuple(self) -> _TimeWindowSortKey:
        return (_WindowSortEnum.TIME_WINDOW, (self.window_start, self.window_end))

    def offset_string(self) -> str:
        return "offset_" + to_human_readable_str(-self.offset) if self.offset.total_seconds() < 0 else ""

    def window_duration_string(self) -> str:
        return to_human_readable_str(self.window_duration)


@attrs.frozen(order=True)
class TimeWindowSeriesSpec(TimeWindowSpec):
    # time_windows represent the tuple of relative time windows - we use a tuple so this can be hashable for secondary key aggregate sorting
    time_windows: Tuple[RelativeTimeWindowSpec]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: TimeWindowSeries) -> "TimeWindowSeriesSpec":
        time_window_specs = ()
        for time_window in proto.time_windows:
            time_window_specs += (RelativeTimeWindowSpec.from_data_proto(time_window),)
        return cls(time_windows=time_window_specs)

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: TimeWindowSeriesArgs) -> "TimeWindowSeriesSpec":
        # If we don't have at least one window in this time range then downstream functions for local development will fail
        time_window_specs = ()
        for i in range(
            proto.series_start.seconds,
            (proto.series_end.seconds - proto.window_duration.seconds) + 1,
            proto.step_size.seconds,
        ):
            time_window_specs += (
                RelativeTimeWindowSpec.from_data_proto(
                    RelativeTimeWindow(
                        window_start=Duration(seconds=i),
                        window_end=Duration(seconds=(i + proto.window_duration.seconds)),
                    )
                ),
            )
        # We need this assertion to ensure that NDD dependencies on this function do not break
        if len(time_window_specs) < 2:
            msg = "Time window series must have at least two windows"
            raise TectonValidationError(msg)
        return cls(time_windows=time_window_specs)

    def to_args_proto(self) -> TimeWindowSeriesArgs:
        return TimeWindowSeriesArgs(
            window_duration=timedelta_to_proto(self.window_duration),
            series_start=timedelta_to_proto(self.window_series_start),
            series_end=timedelta_to_proto(self.window_series_end),
            step_size=timedelta_to_proto(self.step_size),
        )

    def to_data_proto(self) -> TimeWindowSeries:
        return TimeWindowSeries(
            time_windows=[window.to_data_proto() for window in self.time_windows],
        )

    @property
    def window_duration(self) -> datetime.timedelta:
        return self.time_windows[0].window_duration

    @property
    def window_series_start(self) -> datetime.timedelta:
        return self.time_windows[0].window_start

    @property
    def step_size(self) -> datetime.timedelta:
        # At least two windows are guaranteed during validation
        return abs(self.time_windows[1].window_start - self.time_windows[0].window_start)

    @property
    def window_series_end(self) -> datetime.timedelta:
        return self.time_windows[-1].window_end

    def window_series_start_string(self) -> str:
        return to_human_readable_str(abs(self.window_series_start))

    def window_series_end_string(self) -> str:
        return to_human_readable_str(abs(self.window_series_end))

    def step_size_string(self) -> str:
        return to_human_readable_str(self.step_size)

    def window_duration_string(self) -> str:
        return to_human_readable_str(self.window_duration)

    def to_string(self) -> str:
        return f"{self.window_duration_string()}_series_{self.window_series_start_string()}_{self.window_series_end_string()}_{self.step_size_string()}"

    def to_sort_tuple(self) -> _TimeWindowSortKey:
        return (
            _WindowSortEnum.TIME_WINDOW_SERIES,
            (self.window_series_start, self.window_series_end, self.window_duration, self.step_size),
        )


def window_spec_to_window_data_proto(window_spec: TimeWindowSpec) -> TimeWindow:
    if isinstance(window_spec, RelativeTimeWindowSpec):
        return TimeWindow(relative_time_window=window_spec.to_data_proto())
    elif isinstance(window_spec, TimeWindowSeriesSpec):
        return TimeWindow(time_window_series=window_spec.to_data_proto())
    elif isinstance(window_spec, LifetimeWindowSpec):
        return TimeWindow(lifetime_window=window_spec.to_proto())
    else:
        msg = f"Unexpected time window type: {type(window_spec)}"
        raise ValueError(msg)


def create_time_window_spec_from_data_proto(proto: TimeWindow) -> TimeWindowSpec:
    if proto.HasField("time_window_series"):
        return TimeWindowSeriesSpec.from_data_proto(proto.time_window_series)
    elif proto.HasField("relative_time_window"):
        return RelativeTimeWindowSpec.from_data_proto(proto.relative_time_window)
    elif proto.HasField("lifetime_window"):
        return LifetimeWindowSpec()
    else:
        msg = f"Unexpected time window type: {proto}"
        raise ValueError(msg)



import functools
import sys
from dataclasses import dataclass
from typing import Callable
from typing import List
from typing import Optional

import click
import pendulum

from tecton import tecton_context
from tecton._internals.analytics import AnalyticsLogger
from tecton._internals.analytics import StateUpdateEventMetrics
from tecton._internals.analytics import StateUpdateResult
from tecton._internals.utils import cluster_url
from tecton.cli import printer
from tecton.cli import workspace_utils
from tecton_core import conf


analytics = AnalyticsLogger()


class TectonCommand(click.Command):
    # used by integration tests
    skip_config_check = False

    """Base class for Click which implements required_auth and uses_workspace behavior, as well as analytics."""

    def __init__(self, *args, callback, requires_auth=True, uses_workspace=False, **kwargs):
        @functools.wraps(callback)
        @click.pass_context
        def wrapper(ctx, *cbargs, **cbkwargs):
            host = cluster_url()
            cluster_configured = host is not None

            command_names = []
            cur = ctx
            # The top level context is `cli` which we don't want to include.
            while cur:
                command_names.append(cur.command.name)
                cur = cur.parent
            command_names.reverse()

            # TODO(TEC-14547): Move `workspace` param usages to callback function.
            is_create_workspace_command = ctx.parent.info_name == "workspace" and ctx.info_name == "create"
            has_workspace_option = ctx.params.get("workspace", None) is not None
            if has_workspace_option and not is_create_workspace_command:
                workspace_name = ctx.params["workspace"]
                workspace_utils.check_workspace_exists(workspace_name)
                conf.set("TECTON_WORKSPACE", workspace_name)

            # Do not try logging events if cluster has never be configured or if user is trying to log in,
            # otherwise the CLI either won't be able to find the MDS or auth token might have expired
            if cluster_configured:
                if uses_workspace:
                    printer.safe_print(f'Using workspace "{tecton_context.get_current_workspace()}" on cluster {host}')
                start_time = pendulum.now("UTC")
                state_update_event = None
                try:
                    invoke_result = ctx.invoke(callback, *cbargs, **cbkwargs)
                    if isinstance(invoke_result, StateUpdateResult):
                        state_update_event = invoke_result.state_update_event_metrics
                except PermissionError as e:
                    state_update_event = StateUpdateEventMetrics.from_error_message(str(e))
                    printer.safe_print(f"Unable to execute command `{' '.join(command_names)}`: {str(e)}")
                execution_time = pendulum.now("UTC") - start_time
                if requires_auth:
                    if state_update_event:
                        # TODO: Include sub-command?
                        analytics.log_cli_event(command_names[1], execution_time, ctx.params, state_update_event)
                        if state_update_event.error_message:
                            sys.exit(1)
                    else:
                        analytics.log_cli_event(command_names[1], execution_time, ctx.params)
            elif not requires_auth or TectonCommand.skip_config_check:
                # Do not try executing anything besides unauthenticated commnds (`login`, `version`) when cluster hasn't been configured.
                state_update_event = ctx.invoke(callback, *cbargs, **cbkwargs)
            else:
                printer.safe_print(
                    f"`{' '.join(command_names)}` requires authentication. Please authenticate using `tecton login`."
                )
                sys.exit(1)

        super().__init__(*args, callback=wrapper, **kwargs)


class TectonGroup(click.Group):
    """Routes group.command calls to use TectonCommand instead of the base Click command"""

    command_class = TectonCommand


TectonGroup.group_class = TectonGroup


@dataclass
class HiddenValue:
    """Wraps a string and displays it as ****. Used to hide passwords when showing the previous value in prompts."""

    value: str

    def __str__(self):
        return "*" * len(self.value)


class HiddenValueType(click.ParamType):
    name = "TEXT"  # Controls how this is displayed in help text

    def convert(self, value, param=None, ctx=None):
        if value is None:
            return None
        if isinstance(value, HiddenValue):
            return value
        return HiddenValue(value)


def tecton_config_option(
    key: str, param_decls: Optional[List[str]] = None, hide_input: bool = False, **kwargs
) -> Callable:
    """Decorator for a Click option which defaults to a value from tecton_core.conf and prompts if the key is not set.

    :arg key The tecton_core.conf key.
    :arg param_decls The Click param_decls.
    :arg hide_input Whether to hide user input in the prompt, as for passwords. Additionally, if a default is set,
         it will be obfuscated with `*` in the prompt.
    """
    if "prompt" not in kwargs:
        kwargs["prompt"] = True
    lower_key = key.lower().replace("_", "-")
    param_decls = param_decls or [f"--{lower_key}"]

    if hide_input:
        kwargs["type"] = HiddenValueType()
        default = HiddenValueType().convert(conf.get_or_none(key))
    else:

        def default():
            return conf.get_or_none(key)

    def decorator(f):
        @click.option(*param_decls, **kwargs, hide_input=hide_input, default=default)
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            def unwrap(v):
                return v.value if isinstance(v, HiddenValue) else v

            return f(**{k: unwrap(v) for k, v in kwargs.items()})

        return wrapper

    return decorator



import logging
import time
from typing import Optional

import pyspark
from pyspark.sql import SparkSession

from tecton_core.id_helper import IdHelper
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.materialization_utils import df_to_online_store_msg
from tecton_materialization.materialization_utils import set_up_online_store_sink
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark import materialization_plan


logger = logging.getLogger(__name__)


def _get_query_name(is_plan_integration_test: bool = False):
    if is_plan_integration_test:
        return "tecton_osw_sink_plan_integration_test"
    return "tecton_osw_sink"


def _start_stream_job_with_online_store_sink(
    spark: SparkSession, dataframe, materialization_task_params, sink
) -> "pyspark.sql.streaming.StreamingQuery":
    canary_id = materialization_task_params.canary_id if materialization_task_params.HasField("canary_id") else None
    # TODO(amargvela): For SFV add feature timestamp as MATERIALIZED_RAW_DATA_END_TIME column.
    fd = feature_definition_from_task_params(materialization_task_params)

    stream_task_info = materialization_task_params.stream_task_info

    if stream_task_info.HasField("streaming_trigger_interval_override"):
        processing_time = stream_task_info.streaming_trigger_interval_override
    elif fd.is_continuous:
        processing_time = "0 seconds"
    else:
        processing_time = "30 seconds"

    write_df = df_to_online_store_msg(dataframe, fd.id, is_batch=False, is_status=False, canary_id=canary_id)

    logger.info(f"Starting stream write to Tecton Online Store for FV {fd.id}")
    trigger = spark._jvm.org.apache.spark.sql.streaming.Trigger.ProcessingTime(processing_time)
    writer = (
        write_df._jdf.writeStream()
        .queryName(_get_query_name(materialization_task_params.HasField("plan_id")))
        .option(
            "checkpointLocation", f"{stream_task_info.streaming_checkpoint_path}-k"
        )  # append -k to differentiate from Dynamo checkpoint path; keep this in sync with the Canary process.
        .outputMode("update")
        .trigger(trigger)
    )
    # we don't materialize to sinks to not override prod data
    if materialization_task_params.HasField("plan_id"):
        writer = writer.option("startingOffsets", "earliest").format("memory")
    else:
        writer = writer.foreach(sink)
    return writer.start()


def _start_stream_materialization(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    sink,
    job_metadata_client: JobMetadataClient,
) -> "pyspark.sql.streaming.StreamingQuery":
    logger.info(
        f"Starting materialization task {materialization_task_params.materialization_task_id} for feature view {IdHelper.to_string(materialization_task_params.feature_view.feature_view_id)}"
    )

    fd = feature_definition_from_task_params(materialization_task_params)

    plan = materialization_plan.get_stream_materialization_plan(
        spark=spark,
        feature_definition=fd,
    )
    spark_df = plan.online_store_data_frame

    if materialization_task_params.stream_task_info.stream_parameters.stream_handoff_config.enabled:
        _handle_stream_handoff(job_metadata_client)

    online_store_query = _start_stream_job_with_online_store_sink(spark, spark_df, materialization_task_params, sink)

    return online_store_query


def _watch_stream_query(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    stream_query: "pyspark.sql.streaming.StreamingQuery",
    job_metadata_client: JobMetadataClient,
):
    def set_terminated_state(job_metadata: JobMetadata) -> Optional[JobMetadata]:
        new_proto = JobMetadata()
        new_proto.CopyFrom(job_metadata)
        new_proto.spark_execution_info.stream_handoff_synchronization_info.query_cancellation_complete = True
        return new_proto

    def check_memory_output():
        return (
            spark.sql("SELECT * FROM " + _get_query_name(materialization_task_params.HasField("plan_id"))).count() > 0
        )

    stream_params = materialization_task_params.stream_task_info.stream_parameters
    if materialization_task_params.HasField("plan_id"):
        start_time = time.time()
        plan_integration_test_job_timeout = 30 * 60  # 30 minutes in seconds

        while stream_query.isActive():
            if check_memory_output():
                logger.info("Stream query output generated. Stopping stream query.")
                stream_query.stop()
                stream_query.awaitTermination()
                return
            elapsed_time = time.time() - start_time
            if elapsed_time > plan_integration_test_job_timeout:
                logger.info(
                    f"Stream query ran over timeout of {plan_integration_test_job_timeout} seconds. Stopping stream query."
                )
                stream_query.stop()
                stream_query.awaitTermination()
                return
            time.sleep(60)
        # returns immediately or throws exception, given thzat isActive() is false
        stream_query.awaitTermination()
    elif stream_params.stream_handoff_config.enabled:
        while stream_query.isActive():
            job_metadata, _ = job_metadata_client.get()
            # check if the materialization task has been cancelled
            if job_metadata.spark_execution_info.stream_handoff_synchronization_info.query_cancellation_requested:
                logger.info("Stream query cancellation requested. Stopping stream query.")
                try:
                    stream_query.stop()
                    stream_query.awaitTermination()
                finally:
                    logger.info("Query cancellation complete")
                    job_metadata_client.update(set_terminated_state)
                return
            time.sleep(60)
        # returns immediately or throws exception, given that isActive() is false
        stream_query.awaitTermination()
    else:
        stream_query.awaitTermination()


def _handle_stream_handoff(job_metadata_client: JobMetadataClient):
    """
    If stream handoff is enabled, we need to wait for the previous job to finish before starting the next one.
    """

    def set_ready_state(job_metadata: JobMetadata) -> Optional[JobMetadata]:
        new_proto = JobMetadata()
        new_proto.CopyFrom(job_metadata)
        new_proto.spark_execution_info.stream_handoff_synchronization_info.new_cluster_started = True
        return new_proto

    start_time = time.time()
    job_metadata_client.update(set_ready_state)
    logger.info("Using stream handoff; waiting for ready state...")
    job_metadata, _ = job_metadata_client.get()
    while not job_metadata.spark_execution_info.stream_handoff_synchronization_info.stream_query_start_allowed:
        if time.time() - start_time > 3600.0:
            msg = "Timed out waiting for ready state"
            raise Exception(msg)
        time.sleep(1)
        job_metadata, _ = job_metadata_client.get()
    logger.info("Ready state reached; starting streaming query")


def stream_materialize_from_params(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    job_metadata_client: JobMetadataClient,
):
    sink = set_up_online_store_sink(spark, materialization_task_params)
    online_store_sink = _start_stream_materialization(
        spark, materialization_task_params, sink, job_metadata_client=job_metadata_client
    )

    should_publish_stream_metrics = spark.conf.get("spark.tecton.publish_stream_metrics", "true") == "true"

    if should_publish_stream_metrics:
        metricsReportingListener = spark._jvm.com.tecton.onlinestorewriter.MetricsReportingListener(
            materialization_task_params.SerializeToString()
        )
        spark.streams._jsqm.addListener(metricsReportingListener)

    _watch_stream_query(spark, materialization_task_params, online_store_sink, job_metadata_client=job_metadata_client)
    if sink is not None:
        sink.closeGlobalResources()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/dataobs/metric__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)tecton_proto/dataobs/metric__client.proto\x12\x14tecton_proto.dataobs\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xdb\x02\n\x06Metric\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1e\n\nexpression\x18\x02 \x01(\tR\nexpression\x12\x31\n\x06window\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x06window\x12\x35\n\x08interval\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x08interval\x12,\n\x12input_column_names\x18\x05 \x03(\tR\x10inputColumnNames\x12?\n\rcreation_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x63reationTime\x12\x44\n\x10last_update_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0elastUpdateTime\"F\n\x0bMetricValue\x12!\n\x0c\x66\x65\x61ture_name\x18\x01 \x01(\tR\x0b\x66\x65\x61tureName\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value\"\xed\x02\n\x0fMetricDataPoint\x12J\n\x13interval_start_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11intervalStartTime\x12\x46\n\rmetric_values\x18\x02 \x03(\x0b\x32!.tecton_proto.dataobs.MetricValueR\x0cmetricValues\x12\x34\n\x16materialization_run_id\x18\x03 \x01(\tR\x14materializationRunId\x12G\n materialization_task_attempt_url\x18\x04 \x01(\tR\x1dmaterializationTaskAttemptUrl\x12G\n\rmetric_status\x18\x05 \x01(\x0e\x32\".tecton_proto.dataobs.MetricStatusR\x0cmetricStatus\"\x96\x01\n\rFeatureMetric\x12\x41\n\x0bmetric_type\x18\x01 \x01(\x0e\x32 .tecton_proto.dataobs.MetricTypeR\nmetricType\x12!\n\x0c\x66\x65\x61ture_name\x18\x02 \x01(\tR\x0b\x66\x65\x61tureName\x12\x1f\n\x0b\x63olumn_name\x18\x03 \x01(\tR\ncolumnName*\xcb\x01\n\nMetricType\x12\x17\n\x13METRIC_TYPE_UNKNOWN\x10\x00\x12\x12\n\x0e\x43OUNT_DISTINCT\x10\x01\x12\x0e\n\nCOUNT_ROWS\x10\x02\x12\x0f\n\x0b\x43OUNT_NULLS\x10\x03\x12\x0f\n\x0b\x43OUNT_ZEROS\x10\n\x12\r\n\tAVG_VALUE\x10\x04\x12\r\n\tMAX_VALUE\x10\x05\x12\r\n\tMIN_VALUE\x10\x06\x12\x0e\n\nVAR_SAMPLE\x10\x07\x12\x11\n\rSTDDEV_SAMPLE\x10\x08\x12\x0e\n\nAVG_LENGTH\x10\t*\x8b\x01\n\x0cMetricStatus\x12\x19\n\x15METRIC_STATUS_UNKNOWN\x10\x00\x12\x1b\n\x17METRIC_STATUS_AVAILABLE\x10\x01\x12\x1d\n\x19METRIC_STATUS_UNAVAILABLE\x10\x02\x12$\n METRIC_STATUS_NO_MATERIALIZATION\x10\x03\x42\x16\n\x12\x63om.tecton.dataobsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.dataobs.metric__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.tecton.dataobsP\001'
  _METRICTYPE._serialized_start=1076
  _METRICTYPE._serialized_end=1279
  _METRICSTATUS._serialized_start=1282
  _METRICSTATUS._serialized_end=1421
  _METRIC._serialized_start=133
  _METRIC._serialized_end=480
  _METRICVALUE._serialized_start=482
  _METRICVALUE._serialized_end=552
  _METRICDATAPOINT._serialized_start=555
  _METRICDATAPOINT._serialized_end=920
  _FEATUREMETRIC._serialized_start=923
  _FEATUREMETRIC._serialized_end=1073
# @@protoc_insertion_point(module_scope)



import datetime
import enum
import time
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from functools import reduce
from operator import and_
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pendulum
import pyarrow
import pyarrow.compute
import pyarrow.dataset
import pyarrow.fs

from tecton_core import conf
from tecton_core import id_helper
from tecton_core import time_utils
from tecton_core.aws_credentials import assume_role_fetcher
from tecton_core.compute_mode import BatchComputeMode
from tecton_core.errors import TectonInternalError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query_consts import anchor_time
from tecton_core.specs import FeatureTableSpec
from tecton_core.specs import MaterializedFeatureViewSpec
from tecton_proto.common.aws_credentials__client_pb2 import AwsIamRole
from tecton_proto.common.id__client_pb2 import Id
from tecton_proto.data.feature_view__client_pb2 import DeltaOfflineStoreVersion
from tecton_proto.data.feature_view__client_pb2 import OfflineStoreParams
from tecton_proto.data.feature_view__client_pb2 import ParquetOfflineStoreVersion


TIME_PARTITION = "time_partition"
SECONDS_TO_NANOSECONDS = 1000 * 1000 * 1000
CONTINUOUS_PARTITION_SIZE_SECONDS = 86400


class PartitionType(str, enum.Enum):
    DATE_STR = "DateString"
    EPOCH = "Epoch"
    # RAW_TIMESTAMP is only used for Snowflake
    RAW_TIMESTAMP = "RawTimestamp"


class OfflineStoreType(str, enum.Enum):
    PARQUET = "parquet"
    DELTA = "delta"
    SNOWFLAKE = "snowflake"


@attrs.frozen
class TimestampFormats:
    spark_format: str
    python_format: str


@attrs.frozen
class OfflineStorePartitionParams:
    partition_by: str
    partition_type: PartitionType
    partition_interval: pendulum.duration


def partition_size_for_parquet(fd: FeatureDefinitionWrapper) -> pendulum.Duration:
    if fd.offline_store_params is not None:
        return pendulum.Duration(
            seconds=fd.offline_store_params.parquet.time_partition_size.ToTimedelta().total_seconds()
        )
    elif fd.is_continuous:
        return pendulum.Duration(seconds=CONTINUOUS_PARTITION_SIZE_SECONDS)
    else:
        return fd.min_scheduling_interval


def partition_col_for_parquet(fd: FeatureDefinitionWrapper) -> str:
    offline_store_version = (
        fd.offline_store_params.parquet.version
        if fd.offline_store_params is not None
        else ParquetOfflineStoreVersion.PARQUET_OFFLINE_STORE_VERSION_1
    )
    if offline_store_version == ParquetOfflineStoreVersion.PARQUET_OFFLINE_STORE_VERSION_1:
        return TIME_PARTITION if fd.is_continuous else anchor_time()
    elif offline_store_version == ParquetOfflineStoreVersion.PARQUET_OFFLINE_STORE_VERSION_2:
        return TIME_PARTITION
    else:
        msg = "unsupported offline store version"
        raise TectonInternalError(msg)


def partition_size_for_delta(fd: FeatureDefinitionWrapper) -> pendulum.Duration:
    if fd.offline_store_params is not None:
        return pendulum.Duration(
            seconds=fd.offline_store_params.delta.time_partition_size.ToTimedelta().total_seconds()
        )
    else:
        return pendulum.Duration(
            seconds=fd.offline_store_config.delta.time_partition_size.ToTimedelta().total_seconds()
        )


def partition_type_for_delta(offline_store_params: Optional[OfflineStoreParams]) -> PartitionType:
    # TODO(TEC-17350): use this method when reading from the offline store
    if (
        offline_store_params is not None
        and offline_store_params.delta.version == DeltaOfflineStoreVersion.DELTA_OFFLINE_STORE_VERSION_2
    ):
        return PartitionType.EPOCH
    else:
        return PartitionType.DATE_STR


DELTA_SUPPORTED_VERSIONS = [
    DeltaOfflineStoreVersion.DELTA_OFFLINE_STORE_VERSION_1,
    DeltaOfflineStoreVersion.DELTA_OFFLINE_STORE_VERSION_2,
]

PARQUET_SUPPORTED_VERSIONS = [
    ParquetOfflineStoreVersion.PARQUET_OFFLINE_STORE_VERSION_1,
    ParquetOfflineStoreVersion.PARQUET_OFFLINE_STORE_VERSION_2,
]


def _check_supported_offline_store_version(fd: FeatureDefinitionWrapper) -> None:
    if fd.offline_store_params is None:
        return
    if (
        fd.offline_store_params.HasField("delta")
        and fd.offline_store_params.delta.version not in DELTA_SUPPORTED_VERSIONS
    ):
        msg = (
            f"Unsupported offline store version {fd.offline_store_params.delta.version}. Try upgrading your Tecton SDK."
        )
        raise TectonInternalError(msg)
    if (
        fd.offline_store_params.HasField("parquet")
        and fd.offline_store_params.parquet.version not in PARQUET_SUPPORTED_VERSIONS
    ):
        msg = f"Unsupported offline store version {fd.offline_store_params.parquet.version}. Try upgrading your Tecton SDK."
        raise TectonInternalError(msg)


def get_offline_store_type(fd: FeatureDefinitionWrapper) -> OfflineStoreType:
    # TODO(TEC-15800): Update Snowflake FV protos to have snowflake as their store type
    assert isinstance(fd.fv_spec, (MaterializedFeatureViewSpec, FeatureTableSpec))
    if (
        isinstance(fd.fv_spec, MaterializedFeatureViewSpec)
        and fd.fv_spec.batch_compute_mode == BatchComputeMode.SNOWFLAKE
    ):
        return OfflineStoreType.SNOWFLAKE

    store_type = fd.offline_store_config.WhichOneof("store_type")
    if store_type == OfflineStoreType.PARQUET:
        return OfflineStoreType.PARQUET
    elif store_type == OfflineStoreType.DELTA:
        return OfflineStoreType.DELTA
    else:
        msg = f"Unknown offline store type {store_type}"
        raise TectonInternalError(msg)


def get_offline_store_partition_params(feature_definition: FeatureDefinitionWrapper) -> OfflineStorePartitionParams:
    # Examples of how our offline store is partitioned
    ### BWAFV on Delta
    # Partition Column: time_partition
    # Materialized Columns: _anchor_time, [join_keys], [feature_columns]

    ### Continuous SWAFV on Parquet
    # Partition Column: time_partition
    # Materialized Columns: timestamp, _anchor_time, [join_keys], [feature_columns]
    # Note: Very weird that we have a timestamp parquet column here - redundant with _anchor_time

    ### BWAFV on Parquet
    # Partition Column: _anchor_time
    # Materialized Columns: _anchor_time, [join_keys], [feature_columns]
    # !! In this case we need to drop the partition column from the top level columns

    ### BFV on Parquet
    # Partition Column: _anchor_time
    # Materialized Columns: ts, [join_keys], [feature_columns]

    ### Any FV on Snowflake
    # Partition Column: none
    # Materialized Columns: ts, [join_keys], [feature_columns]

    _check_supported_offline_store_version(feature_definition)
    store_type = get_offline_store_type(feature_definition)

    if store_type == OfflineStoreType.SNOWFLAKE:
        partition_by = feature_definition.time_key
        partition_type = PartitionType.RAW_TIMESTAMP
        partition_interval = pendulum.Duration(seconds=0)
    elif store_type == OfflineStoreType.DELTA:
        partition_by = TIME_PARTITION
        partition_type = PartitionType.DATE_STR
        partition_interval = partition_size_for_delta(feature_definition)
    elif store_type == OfflineStoreType.PARQUET:
        partition_by = partition_col_for_parquet(feature_definition)
        partition_type = PartitionType.EPOCH
        partition_interval = partition_size_for_parquet(feature_definition)
    else:
        msg = "Unexpected offline store config"
        raise Exception(msg)
    return OfflineStorePartitionParams(partition_by, partition_type, partition_interval)


def timestamp_to_partition_date_str(timestamp: pendulum.DateTime, partition_params: OfflineStorePartitionParams) -> str:
    partition_interval_timedelta = partition_params.partition_interval.as_timedelta()
    aligned_time = time_utils.align_time_downwards(timestamp, partition_interval_timedelta)
    partition_format = timestamp_formats(partition_interval_timedelta).python_format
    return aligned_time.strftime(partition_format)


def timestamp_to_partition_epoch(
    timestamp: pendulum.DateTime,
    partition_params: OfflineStorePartitionParams,
    feature_store_format_version: int,
) -> int:
    aligned_time = time_utils.align_time_downwards(timestamp, partition_params.partition_interval.as_timedelta())
    # align_time_downwards returns the time without tzinfo. convert_timestamp_for_version calls timestamp() which
    # treats naive datetime instances as local time. This can cause an issue if local time is not in UTC.
    aligned_time = aligned_time.replace(tzinfo=datetime.timezone.utc)
    return time_utils.convert_timestamp_for_version(aligned_time, feature_store_format_version)


def window_size_seconds(window: Union[timedelta, pendulum.Duration]) -> int:
    if isinstance(window, pendulum.Duration):
        window = window.as_timedelta()
    if window % timedelta(seconds=1) != timedelta(0):
        msg = f"partition_size is not a round number of seconds: {window}"
        raise AssertionError(msg)
    return int(window.total_seconds())


def timestamp_formats(partition_size: timedelta) -> TimestampFormats:
    if partition_size % timedelta(days=1) == timedelta(0):
        return TimestampFormats(spark_format="yyyy-MM-dd", python_format="%Y-%m-%d")
    else:
        return TimestampFormats(spark_format="yyyy-MM-dd-HH:mm:ss", python_format="%Y-%m-%d-%H:%M:%S")


def datetime_to_partition_str(dt: datetime, partition_size: timedelta) -> str:
    partition_format = timestamp_formats(partition_size).python_format
    return dt.strftime(partition_format)


@dataclass
class OfflineStoreReaderParams:
    delta_table_uri: str


@dataclass
class JoinKeyBoundaries:
    min: Union[int, str]
    max: Union[int, str]


class OfflineStoreReader(ABC):
    @abstractmethod
    def read(
        self,
        partition_time_limits: Optional[pendulum.Period] = None,
        join_keys_limits: Optional[Dict[str, JoinKeyBoundaries]] = None,
    ) -> "pyarrow.dataset.Dataset":
        raise NotImplementedError


class S3OptionsSource(enum.Enum):
    LOCAL = 1
    MDS = 2


@dataclass
class S3Options:
    access_key_id: str
    secret_access_key: str
    session_token: Optional[str]
    options_source: S3OptionsSource
    region: Optional[str] = None

    def to_dict(self):
        opts = {
            "AWS_ACCESS_KEY_ID": self.access_key_id,
            "AWS_SECRET_ACCESS_KEY": self.secret_access_key,
        }
        if self.session_token:
            opts["AWS_SESSION_TOKEN"] = self.session_token
        if self.region:
            opts["AWS_REGION"] = self.region
        return opts


class OfflineStoreOptionsProvider:
    def get_s3_options_for_feature_view(self, feature_view_id: Id) -> Optional[S3Options]:
        return

    def get_s3_options_for_data_source(self, data_source_id: Id) -> Optional[S3Options]:
        return


class BotoOfflineStoreOptionsProvider(OfflineStoreOptionsProvider):
    """deltalake's built-in S3 auth is implemented by a Rust library which doesn't support the full range of AWS
    auth configurations supported by boto, which can be surprising to users used to boto's options. This class
    following does a fully-featured auth using boto instead of depending on Rust.

    See also https://github.com/delta-io/delta-rs/issues/855
    """

    @staticmethod
    def static_options(assume_role: Optional[AwsIamRole] = None) -> Optional[S3Options]:
        import boto3

        if assume_role:
            return BotoOfflineStoreOptionsProvider._cross_account_options(assume_role)

        session = boto3.Session()
        credentials = session.get_credentials()
        if credentials is None:
            return None
        current_credentials = credentials.get_frozen_credentials()
        return S3Options(
            access_key_id=current_credentials.access_key,
            secret_access_key=current_credentials.secret_key,
            session_token=current_credentials.token,
            options_source=S3OptionsSource.LOCAL,
            region=conf.get_or_none("CLUSTER_REGION"),
        )

    @staticmethod
    def _cross_account_options(assume_role: AwsIamRole) -> Optional[S3Options]:
        credentials = assume_role_fetcher(
            assume_role, session_name=f"CrossAccountSession-{datetime.datetime.now().strftime('%Y%m%d%H%M%S')}"
        ).fetch_credentials()
        return S3Options(
            access_key_id=credentials["access_key"],
            secret_access_key=credentials["secret_key"],
            session_token=credentials["token"],
            options_source=S3OptionsSource.LOCAL,
            region=conf.get_or_none("CLUSTER_REGION"),
        )

    def get_s3_options_for_feature_view(self, feature_view_id: Id) -> Optional[S3Options]:
        return BotoOfflineStoreOptionsProvider.static_options()

    def get_s3_options_for_data_source(self, data_source_id: Id) -> Optional[S3Options]:
        return BotoOfflineStoreOptionsProvider.static_options()


DEFAULT_OPTIONS_PROVIDERS = [BotoOfflineStoreOptionsProvider()]


def get_s3_options_for_fd(
    fd: FeatureDefinitionWrapper, options_providers: Iterable[OfflineStoreOptionsProvider]
) -> S3Options:
    fvid = id_helper.IdHelper.from_string(fd.id)
    options = next(
        filter(lambda o: o is not None, (p.get_s3_options_for_feature_view(fvid) for p in options_providers)), None
    )
    if options is None:
        msg = f"Unable to retrieve S3 store credentials for feature view {fd.name}."
        raise ValueError(msg)

    return options


def patch_timestamps_in_arrow_schema(schema: pyarrow.Schema) -> pyarrow.Schema:
    """Patches schema to replace tz-naive columns with UTC timestamps"""
    for idx, field in enumerate(schema):
        if isinstance(field.type, pyarrow.TimestampType) and field.type.tz is None:
            schema = schema.set(idx, pyarrow.field(field.name, pyarrow.timestamp(field.type.unit, "UTC")))
    return schema


class DeltaReader(OfflineStoreReader):
    def __init__(
        self,
        params: OfflineStoreReaderParams,
        fd: FeatureDefinitionWrapper,
        options_providers: Iterable[OfflineStoreOptionsProvider],
    ) -> None:
        self._params = params
        self._partition_size = partition_size_for_delta(fd).as_timedelta()
        self._fd = fd
        self._options_providers = options_providers

    def _storage_options(self) -> Optional[S3Options]:
        if self._params.delta_table_uri.startswith("s3"):
            return get_s3_options_for_fd(self._fd, self._options_providers)

        return None

    @property
    def _table(self):
        # TODO(TEC-16757): Move import to top of file.
        try:
            from deltalake import DeltaTable
        except ImportError:
            msg = (
                "Couldn't initialize Rift compute. "
                "To use Rift install all Rift dependencies first by executing `pip install tecton[rift]`."
            )
            raise RuntimeError(msg)

        options = self._storage_options()
        options_dict = {}
        if options:
            options_dict = options.to_dict()

        # Sometimes the backend credentials we receive are not considered valid by S3 for some time. We want to silently
        # retry until they become valid when this happens.
        #
        # This is done in a rather crude way here. It would be better to test the credentials out using boto3 which is
        # faster/gives us more granular errors, but inexplicably there are at least some cases where the credentials
        # work with boto3 but then later fail when passed to Delta.
        saved_error = None
        for _ in range(20):
            try:
                return DeltaTable(table_uri=self._params.delta_table_uri, storage_options=options_dict)
            except OSError as e:
                saved_error = e
                time.sleep(0.1)
        msg = "Failed to read from S3"
        if not options or S3OptionsSource.LOCAL == options.options_source:
            msg = "Failed to read from S3 using user-provided AWS credentials"
        elif S3OptionsSource.MDS == options.options_source:
            msg = "Failed to read from S3 using Tecton-provided temporary credentials"
        raise TimeoutError(msg) from saved_error

    def _partition_filter(self, partition_time_limits: Optional[pendulum.Period] = None) -> List[Tuple[str, str, str]]:
        if partition_time_limits is None or not self._partition_size:
            return

        filters = []
        aligned_start_time = time_utils.align_time_downwards(partition_time_limits.start, self._partition_size)
        start_partition = datetime_to_partition_str(aligned_start_time, self._partition_size)
        filters.append((TIME_PARTITION, ">=", start_partition))

        aligned_end_time = time_utils.align_time_downwards(partition_time_limits.end, self._partition_size)
        end_partition = datetime_to_partition_str(aligned_end_time, self._partition_size)
        filters.append((TIME_PARTITION, "<=", end_partition))
        return filters

    def _join_keys_filter(
        self, join_keys_limits: Optional[Dict[str, JoinKeyBoundaries]] = None
    ) -> Optional[pyarrow.dataset.Expression]:
        if not join_keys_limits:
            return

        conditions = []
        for join_key, boundaries in join_keys_limits.items():
            conditions.append(
                (pyarrow.compute.field(join_key) >= boundaries.min)
                & (pyarrow.compute.field(join_key) <= boundaries.max)
            )

        return reduce(and_, conditions)

    def read(
        self,
        partition_time_limits: Optional[pendulum.Period] = None,
        join_keys_limits: Optional[Dict[str, JoinKeyBoundaries]] = None,
    ) -> "pyarrow.dataset.Dataset":
        ds = self._table.to_pyarrow_dataset(partitions=self._partition_filter(partition_time_limits))
        fragments = ds.get_fragments(filter=self._join_keys_filter(join_keys_limits))
        schema = patch_timestamps_in_arrow_schema(ds.schema)
        return pyarrow.dataset.FileSystemDataset(fragments, schema, ds.format, ds.filesystem)


class ParquetReader(OfflineStoreReader):
    def __init__(
        self,
        fd: FeatureDefinitionWrapper,
        options_providers: Iterable[OfflineStoreOptionsProvider],
    ) -> None:
        self._fd = fd
        self._options_providers = options_providers
        self._partition_params = get_offline_store_partition_params(self._fd)

    def _partition_column_type(self):
        return pyarrow.int64() if self._partition_params.partition_type == PartitionType.EPOCH else pyarrow.string()

    def _filesystem_and_uri(self) -> Tuple[pyarrow.fs.FileSystem, str]:
        uri = self._fd.materialized_data_path
        fs, path = pyarrow.fs.FileSystem.from_uri(uri)

        if isinstance(fs, pyarrow.fs.S3FileSystem):
            options = get_s3_options_for_fd(self._fd, self._options_providers)

            return (
                pyarrow.fs.S3FileSystem(
                    access_key=options.access_key_id,
                    secret_key=options.secret_access_key,
                    session_token=options.session_token,
                ),
                path,
            )

        return fs, path

    def _filter(self, partition_time_limits: Optional[pendulum.Period]) -> pyarrow.dataset.Expression:
        if not partition_time_limits:
            return

        start_partition = timestamp_to_partition_epoch(
            partition_time_limits.start,
            self._partition_params,
            self._fd.get_feature_store_format_version,
        )
        end_partition = timestamp_to_partition_epoch(
            partition_time_limits.end,
            self._partition_params,
            self._fd.get_feature_store_format_version,
        )
        column = pyarrow.compute.field(self._partition_params.partition_by)
        return (column >= start_partition) & (column <= end_partition)

    def read(
        self,
        partition_time_limits: Optional[pendulum.Period] = None,
        join_keys_limits: Optional[Dict[str, JoinKeyBoundaries]] = None,
    ) -> "pyarrow.dataset.Dataset":
        partitioning = pyarrow.dataset.partitioning(
            pyarrow.schema([(self._partition_params.partition_by, self._partition_column_type())]), flavor="hive"
        )
        fs, uri = self._filesystem_and_uri()

        # We had to use an explicit dataset creation (via FileSystemDatasetFactory)
        # to override default `selector_ignore_prefixes` parameter.
        # The default behaviour is to ignore directories/files starting with ["_", "."],
        # and we use column `_anchor_time` as a partition column.
        selector = pyarrow.fs.FileSelector(uri, recursive=True)
        file_format = pyarrow.dataset.ParquetFileFormat()
        options = pyarrow.dataset.FileSystemFactoryOptions(
            partitioning=partitioning,
            selector_ignore_prefixes=[".", "_SUCCESS", "_committed", "_started"],
        )
        return pyarrow.dataset.FileSystemDatasetFactory(fs, selector, file_format, options).finish()



from typing import List
from typing import Optional
from typing import Set

from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonValidationError
from tecton_core.schema import Schema
from tecton_proto.common import schema__client_pb2 as schema_pb2


# TODO(njoung): Enforce some consistency between callable functions and instantiated objects
# in the exports of this file


# Generic
def INTERNAL_ERROR(message):
    return TectonInternalError(
        f"We seem to have encountered an error. Please contact support for assistance. Error details: {message}"
    )


def MDS_INACCESSIBLE(host_port):
    return TectonInternalError(
        f"Failed to connect to Tecton at {host_port}, please check your connectivity or contact support"
    )


def VALIDATION_ERROR_FROM_MDS(message, trace_id: Optional[str] = None):
    suffix = f", trace ID: {trace_id}" if trace_id else ""
    return TectonValidationError(f"{message}{suffix}")


def DEPENDENT_FEATURE_VIEWS_REQUIRE_SCHEMAS(method_name, name, feature_views_missing_schemas):
    msg = (
        f"Method {method_name} requires `features` parameter: Feature Service {name} has dependent Feature View(s) "
        f"{feature_views_missing_schemas} which did not specify schema or features parameter."
    )
    return TectonValidationError(msg)


def UNSUPPORTED_OPERATION(op, reason):
    return TectonValidationError(f"Operation '{op}' is not supported: {reason}")


def INVALID_SPINE_TIME_KEY_TYPE_SPARK(t):
    return TectonValidationError(
        f"Invalid type of timestamp_key column in the input events dataframe. Expected TimestampType, got {t}"
    )


INVALID_NULL_SPINE_TIME_KEY = TectonValidationError(
    "Unable to infer the time range of the events dataframe. This typically occurs when all the timestamps in the dataframe are null."
)


def INVALID_SPINE_TIME_KEY_TYPE_PANDAS(t):
    return TectonValidationError(
        f"Invalid type of timestamp_key column in the given events dataframe. Expected datetime, got {t}"
    )


def MISSING_SPINE_COLUMN(param, col, existing_cols):
    return TectonValidationError(
        f"{param} column is missing from the events dataframe. Expected to find '{col}' among available columns: '{', '.join(existing_cols)}'."
    )


def MISSING_REQUEST_DATA_IN_SPINE(key, existing_cols):
    return TectonValidationError(
        f"Request context key '{key}' not found in the events dataframe schema. Expected to find '{key}' among available columns: '{', '.join(existing_cols)}'."
    )


def NONEXISTENT_WORKSPACE(name, workspaces):
    return TectonValidationError(f'Workspace "{name}" not found. Available workspaces: {workspaces}')


def INCORRECT_MATERIALIZATION_ENABLED_FLAG(user_set_bool, server_side_bool):
    return TectonValidationError(
        f"'is_live={user_set_bool}' argument does not match the value on the server: {server_side_bool}"
    )


def UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE(op):
    return TectonValidationError(f"Operation '{op}' is not supported in a development workspace")


def INVALID_JOIN_KEYS_TYPE(t):
    return TectonValidationError(f"Invalid type for join_keys. Expected Dict[str, Union[int, str, bytes]], got {t}")


def INVALID_REQUEST_DATA_TYPE(t):
    return TectonValidationError(
        f"Invalid type for request_data. Expected Dict[str, Union[int, str, bytes, float]], got {t}"
    )


def INVALID_REQUEST_CONTEXT_TYPE(t):
    return TectonValidationError(
        f"Invalid type for request_context_map. Expected Dict[str, Union[int, str, bytes, float]], got {t}"
    )


def INVALID_INDIVIDUAL_JOIN_KEY_TYPE(key: str, type_str: str):
    return TectonValidationError(
        f"Invalid type for join_key '{key}'. Expected either type int, str, or bytes, got {type_str}"
    )


def EMPTY_ARGUMENT(argument: str):
    return TectonValidationError(f"Argument '{argument}' can not be empty.")


def EMPTY_ELEMENT_IN_ARGUMENT(argument: str):
    return TectonValidationError(f"Argument '{argument}' can not have an empty element.")


def DUPLICATED_ELEMENTS_IN_ARGUMENT(argument: str):
    return TectonValidationError(f"Argument '{argument}' can not have duplicated elements.")


def DATA_SOURCE_HAS_NO_BATCH_CONFIG(data_source: str):
    return TectonValidationError(
        f"Cannot run get_dataframe on locally defined Data Source '{data_source}' because it does not have a batch_config"
    )


def FEATURE_VIEW_HAS_NO_BATCH_SOURCE(feature_view: str):
    return TectonValidationError(
        f"Cannot run get_historical_features with from_source=True for Feature View {feature_view} because it depends on a Data Source which does not have a batch config set. Please retry with from_source=False"
    )


def FEATURE_VIEW_HAS_NO_STREAM_SOURCE(feature_view: str):
    return TectonValidationError(
        f"Cannot run run_stream on Feature View {feature_view} because it does not have a Stream Source"
    )


def UNKNOWN_REQUEST_CONTEXT_KEY(keys, key):
    return TectonValidationError(f"Unknown request context key '{key}', expected one of: {keys}")


def FV_TIME_KEY_MISSING(fv_name):
    return TectonValidationError(f"Argument 'timestamp_key' is required for the feature definition '{fv_name}'")


def FV_NO_MATERIALIZED_DATA(fv_name):
    return TectonValidationError(
        f"Feature definition '{fv_name}' doesn't have any materialized data. Materialization jobs may not have updated the offline feature store yet. Please monitor using materialization_status() or use from_source=True to compute from source data."
    )


def ODFV_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE(odfv_name, workspace):
    return TectonValidationError(
        f"On-Demand Feature View {odfv_name} is in workspace {workspace}, which is a development workspace (does not have materialization enabled). Please use from_source=True when retrieving features or alternatively use a live workspace and configure offline materialization for all dependent Feature Views."
    )


def FD_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE(fd_name, workspace):
    return TectonValidationError(
        f"Feature Definition {fd_name} is in workspace {workspace}, which is a development workspace (does not have materialization enabled). Please use from_source=True when getting features (not applicable for Feature Tables) or alternatively configure offline materialization for this Feature Definition in a live workspace."
    )


def FEATURE_TABLE_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE(ft_name, workspace):
    return TectonValidationError(
        f"Feature Table {ft_name} is in workspace {workspace}, which is a development workspace (does not have materialization enabled). Please apply this Feature Table to a live workspace and ingest some features before using with get_historical_features()."
    )


def FEATURE_TABLE_GET_ONLINE_FEATURES_FROM_DEVELOPMENT_WORKSPACE(ft_name, workspace):
    return TectonValidationError(
        f"Feature Table {ft_name} is in workspace {workspace}, which is a development workspace (does not have materialization enabled). Please apply this Feature Table to a live workspace and ingest some features before using with get_online_features()."
    )


def FEATURE_TABLE_GET_MATERIALIZED_FEATURES_OFFLINE_FALSE(ft_name):
    return TectonValidationError(
        f"Feature Table {ft_name} does not have offline materialization enabled, i.e. offline=True. Cannot retrieve offline feature if offline materializaiton is not enabled."
    )


def FD_GET_MATERIALIZED_FEATURES_FROM_LOCAL_OBJECT(fv_name, fco_name):
    return TectonValidationError(
        f"{fco_name} {fv_name} is defined locally, i.e. it has not been applied to a Tecton workspace. In order to force fetching data from the Offline store (i.e. from_source=False) this Feature View must be applied to a Live workspace and have materialization enabled (i.e. offline=True)."
    )


def UNSUPPORTED_SAVE_IN_READ_API_WITH_LOCAL_OBJECT(fs_name):
    return TectonValidationError(
        f"{fs_name} is defined locally, i.e. it has not been applied to a Tecton workspace. In order to save features (i.e. save=True) this Feature Service must be applied to a workspace and have materialization enabled (i.e. offline=True)."
    )


def FD_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE_GFD(fv_name, workspace):
    return TectonValidationError(
        f"Feature View {fv_name} is in workspace {workspace}, which is a development workspace (does not have materialization enabled). In order to force fetching data from the Offline store (i.e. from_source=False) this Feature View must be applied to a Live workspace and have materialization enabled (i.e. offline=True)."
    )


def FV_WITH_INC_BACKFILLS_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE(fv_name, workspace):
    return TectonValidationError(
        f"Feature view {fv_name} uses incremental backfills and is in workspace {workspace}, which is a development workspace (does not have materialization enabled). "
        + "Computing features from source is not supported for Batch Feature Views with incremental_backfills set to True. "
        + "Enable offline materialization for this feature view in a live workspace to use `get_historical_features()`, or use `run()` to test this feature view without materializing data."
    )


def FV_WITH_INC_BACKFILLS_GET_MATERIALIZED_FEATURES_IN_LOCAL_MODE(fv_name):
    return TectonValidationError(
        f"Feature view {fv_name} uses incremental backfills and is locally defined which means materialization is not enabled."
        + "Computing features from source is not supported for Batch Feature Views with incremental_backfills set to True. "
        + "Apply and enable offline materialization for this feature view in a live workspace to use `get_historical_features()`, or use `run()` to test this feature view locally in a notebook."
    )


def FV_WITH_INC_BACKFILLS_GET_MATERIALIZED_FEATURES_MOCK_DATA(
    fv_name: str, method_name: str = "get_historical_features"
):
    return TectonValidationError(
        f"Feature view {fv_name} uses incremental backfills and is locally defined which means materialization is not enabled."
        + "Computing features from mock data is not supported for Batch Feature Views with incremental_backfills set to True."
        + f"Apply and enable offline materialization for this feature view in a live workspace to use `{method_name}()` or use `run_transformation()` to test this Feature View."
    )


def FD_GET_FEATURES_MATERIALIZATION_DISABLED(fd_name):
    return TectonValidationError(
        f"Feature View {fd_name} does not have offline materialization turned on. In order to force fetching data from the Offline store (i.e. from_source=False) this Feature View must be applied to a Live workspace and have materialization enabled (i.e. offline=True)."
    )


def FV_GET_FEATURES_MATERIALIZATION_DISABLED_GFD(fv_name):
    return TectonValidationError(
        f"Feature View {fv_name} does not have offline materialization turned on. Try calling this function with 'use_materialized_data=False' or alternatively configure offline materialization for this Feature View."
    )


# DataSources
DS_STREAM_PREVIEW_ON_NON_STREAM = TectonValidationError("'start_stream_preview' called on non-streaming data source")

DS_DATAFRAME_NO_TIMESTAMP = TectonValidationError(
    "Cannot find timestamp column for this data source. Please call 'get_dataframe' without parameters 'start_time' or 'end_time'."
)

DS_RAW_DATAFRAME_NO_TIMESTAMP_FILTER = TectonValidationError(
    "The method 'get_dataframe()' cannot filter on timestamps when 'apply_translator' is False. "
    "'start_time' and 'end_time' must be None."
)

DS_INCORRECT_SUPPORTS_TIME_FILTERING = TectonValidationError(
    "Cannot filter on timestamps when supports_time_filtering on data source is False. "
    "'start_time' and 'end_time' must be None."
)


def FS_INTERNAL_ERROR(message):
    return TectonInternalError(f"Online feature request error: {message}")


FS_GET_FEATURE_VECTOR_REQUIRED_ARGS = TectonValidationError(
    "get_feature_vector requires at least one of join_keys or request_context_map"
)

FS_API_KEY_MISSING = TectonValidationError(
    "API key is required for online feature requests, but was not found in the environment. Please generate a key and set TECTON_API_KEY "
    + "using https://docs.tecton.ai/docs/reading-feature-data/reading-feature-data-for-inference"
)


def FV_INVALID_MOCK_SOURCES(mock_sources_keys: List[str], fv_params: List[str]):
    return TectonValidationError(
        f"Mock sources {mock_sources_keys} do not match the Feature View's input parameters {fv_params}"
    )


def FV_INVALID_MOCK_INPUTS(mock_inputs: List[str], inputs: List[str]):
    return TectonValidationError(f"Mock input {mock_inputs} do not match FeatureView's inputs {inputs}")


def UNDEFINED_REQUEST_SOURCE_INPUT(undefined_inputs: List[str], expected_inputs: List[str]):
    return TectonValidationError(
        f"The provided request source data contains keys not defined in the request source schema. Extraneous keys: {undefined_inputs}. Expected keys: {expected_inputs}."
    )


def FV_INVALID_MOCK_INPUTS_NUM_ROWS(num_rows: List[int]):
    return TectonValidationError(
        f"Number of rows are not equal across all mock_inputs. Number of rows found are: {str(num_rows)}."
    )


def FV_UNSUPPORTED_ARG(invalid_arg_name: str):
    return TectonValidationError(f"Argument '{invalid_arg_name}' is not supported for this FeatureView type.")


def FV_INVALID_ARG_VALUE(arg_name: str, value: str, expected: str):
    return TectonValidationError(f"Invalid argument value '{arg_name}={value}', supported value(s): '{expected}'")


def FV_INVALID_ARG_COMBO(arg_names: List[str]):
    return TectonValidationError(f"Invalid argument combinations; {str(arg_names)} cannot be used together.")


def FT_UNABLE_TO_ACCESS_SOURCE_DATA(fv_name):
    return TectonValidationError(
        f"The source data for FeatureTable {fv_name} does not exist. Please use from_source=False when calling this function."
    )


def NO_SCHEMA_FROM_SOURCE_FALSE(fv_name, method_name):
    return TectonValidationError(
        f"Feature view {fv_name} does not have a `features` parameter, and `from_source` is set to False. When calling `{method_name}`, either the `features` param must be used or `from_source` must be set to `True`. Please use the `features` parameter, or set `from_source=True`."
    )


class InvalidTransformationMode(TectonValidationError):
    def __init__(self, name: str, got: str, allowed_modes: List[str]):
        super().__init__(f"Mode for '{name}' is '{got}', must be one of: {', '.join(allowed_modes)}")


class InvalidConstantType(TectonValidationError):
    def __init__(self, value, allowed_types):
        allowed_types = [str(allowed_type) for allowed_type in allowed_types]
        super().__init__(
            f"Tecton const value '{value}' must have one of the following types: {', '.join(allowed_types)}"
        )


class InvalidTransformInvocation(TectonValidationError):
    def __init__(self, transformation_name: str, got: str):
        super().__init__(
            f"Allowed arguments for Transformation '{transformation_name}' are: "
            f"tecton.const, tecton.materialization_context, transformations, and DataSource inputs. Got: '{got}'"
        )


# Dataset
DATASET_SPINE_COLUMNS_NOT_SET = TectonValidationError(
    "Cannot retrieve the events dataframe when Dataset was created without one."
)

UNSUPPORTED_FETCH_AS_PANDAS_AVRO = TectonValidationError(
    "Logged datasets require spark. Please use `to_spark()` to fetch."
)


def INVALID_DATASET_PATH(path: str):
    return TectonValidationError(f"Dataset storage location must be an s3 path or a local directory. path={path}")


# Feature Retrevial
def GET_HISTORICAL_FEATURES_WRONG_PARAMS(params: List[str], if_statement: str):
    return TectonValidationError("Cannot provide parameters " + ", ".join(params) + f" if {if_statement}")


GET_ONLINE_FEATURES_REQUIRED_ARGS = TectonValidationError(
    "get_online_features requires at least one of join_keys or request_data to be set."
)

GET_ONLINE_FEATURES_ODFV_JOIN_KEYS = TectonValidationError(
    "get_online_features requires the 'join_keys' argument for this On-Demand Feature View since it has other Feature Views as inputs"
)

GET_ONLINE_FEATURES_FS_JOIN_KEYS = TectonValidationError(
    "get_online_features requires the 'join_keys' argument for this Feature Service"
)

GET_FEATURE_VECTOR_FS_JOIN_KEYS = TectonValidationError(
    "get_feature_vector requires the 'join_keys' argument for this Feature Service"
)

FS_GET_ONLINE_FEATURES_REQUIRED_ARGS = TectonValidationError(
    "get_online_features requires at least one of join_keys or request_data"
)


def GET_ONLINE_FEATURES_MISSING_REQUEST_KEY(keys: Set[str]):
    return TectonValidationError("The following required keys are missing in request_data: " + ", ".join(keys))


def GET_FEATURE_VECTOR_MISSING_REQUEST_KEY(keys: Set[str]):
    return TectonValidationError("The following required keys are missing in request_context_map: " + ", ".join(keys))


def GET_ONLINE_FEATURES_FS_NO_REQUEST_DATA(keys: List[str]):
    return TectonValidationError(
        "get_online_features requires the 'request_data' argument for this Feature Service since it contains an On-Demand Feature View. "
        + "Expected the following request data keys: "
        + ", ".join(keys)
    )


def GET_FEATURE_VECTOR_FS_NO_REQUEST_DATA(keys: List[str]):
    return TectonValidationError(
        "get_feature_vector requires the 'request_context_map' argument for this Feature Service since it contains an On-Demand Feature View. "
        + "Expected the following request context keys: "
        + ", ".join(keys)
    )


def GET_ONLINE_FEATURES_FV_NO_REQUEST_DATA(keys: List[str]):
    return TectonValidationError(
        "get_online_features requires the 'request_data' argument for On-Demand Feature Views. Expected the following request context keys: "
        + ", ".join(keys)
    )


def GET_FEATURE_VECTOR_FV_NO_REQUEST_DATA(keys: List[str]):
    return TectonValidationError(
        "get_feature_vector requires the 'request_context_map' argument for On-Demand Feature Views. Expected the following request context keys: "
        + ", ".join(keys)
    )


def ODFV_WITH_FT_INPUT_DEV_WORKSPACE(ft_name):
    return TectonValidationError(
        f"This On-Demand Feature View has a Feature Table, {ft_name}, as an input. Feature Table feature retrieval cannot be used in a dev workspace. Please apply the Feature Table to a live workspace to retrieve features."
    )


def ODFV_WITH_FT_INPUT_LOCAL_MODE(ft_name):
    return TectonValidationError(
        f"This On-Demand Feature View has a Feature Table, {ft_name}, as an input. Feature Table feature retrieval cannot be used when the Feature Table is locally defined. Please apply the Feature Table to a live workspace to retrieve features."
    )


def ODFV_WITH_FT_INPUT_FROM_SOURCE(ft_name):
    return TectonValidationError(
        f"This On-Demand Feature View has a Feature Table, {ft_name}, as an input. Feature Table features must be retrieved from the offline store, i.e. with from_source=False."
    )


def ODFV_WITH_UNMATERIALIZED_FV_INPUT_FROM_SOURCE_FALSE(fv_name):
    return TectonValidationError(
        f"This On-Demand Feature View has a Feature View, {fv_name}, as an input, which does not have materialization enabled (offline=False). Either retrieve features with from_source=True (not applicable for Feature Tables) or enable offline materialization for the input feature views."
    )


def LOCAL_ODFV_WITH_DEV_WORKSPACE_FV_INPUT_FROM_SOURCE_FALSE(fv_name):
    return TectonValidationError(
        f"This On-Demand Feature View has a Feature View, {fv_name}, as an input, which belongs to a dev workspace and therefore does not have materialization enabled. Either retrieve features with from_source=True (not applicable for Feature Tables) or apply this Feature View to a live workspace with offline=True."
    )


def LOCAL_ODFV_WITH_LOCAL_FV_INPUT_FROM_SOURCE_FALSE(fv_name):
    return TectonValidationError(
        f"This On-Demand Feature View has a Feature View, {fv_name}, as an input, that is locally defined and therefore does not have materialization enabled. Either retrieve features with from_source=True or apply this Feature View to a live workspace with offline=True."
    )


FROM_SOURCE_WITH_FT = TectonValidationError(
    "Computing features from source is not supported for Feature Tables. Try calling this method with from_source=False."
)

USE_MATERIALIZED_DATA_WITH_FT = TectonValidationError(
    "Computing features from source is not supported for Feature Tables. Try calling this method with use_materialized_data=True."
)

FS_WITH_FT_DEVELOPMENT_WORKSPACE = TectonValidationError(
    "This Feature Service contains a Feature Table and fetching historical features for Feature Tables is not supported in a development workspace. This method is only supported in live workspaces."
)

FV_WITH_FT_DEVELOPMENT_WORKSPACE = TectonValidationError(
    "This Feature View has a Feature Table input and fetching historical features for Feature Tables is not supported in a development workspace. This method is only supported in live workspaces."
)

# Backfill Config Validation
BFC_MODE_SINGLE_REQUIRED_FEATURE_END_TIME_WHEN_START_TIME_SET = TectonValidationError(
    "feature_end_time is required when feature_start_time is set, for a FeatureView with "
    + "single_batch_schedule_interval_per_job backfill mode."
)

BFC_MODE_SINGLE_INVALID_FEATURE_TIME_RANGE = TectonValidationError(
    "Run with single_batch_schedule_interval_per_job backfill mode only supports time range equal to batch_schedule"
)


def INCORRECT_KEYS(keys, join_keys):
    return TectonValidationError(
        f"Requested keys to be deleted ({keys}) do not match the expected join keys ({join_keys})."
    )


NO_STORE_SELECTED = TectonValidationError("One of online or offline store must be selected.")


def TOO_MANY_KEYS(max_keys: int):
    return TectonValidationError(f"Max number of keys to be deleted is {max_keys}.")


OFFLINE_STORE_NOT_SUPPORTED = TectonValidationError(
    "Only DeltaLake is supported for entity deletion in offline feature stores."
)

FV_UNSUPPORTED_AGGREGATION = TectonValidationError(
    "Argument 'aggregation_level' is not supported for Feature Views with `aggregations` not specified."
)

RUN_API_PARTIAL_LEVEL_UNSUPPORTED_FOR_COMPACTION = TectonValidationError(
    "aggregation_level='partial' is only supported for Aggregate Feature Views with Compaction Disabled. Use aggregation_level='full' or aggregation_level='disabled' instead."
)


def INVALID_JOIN_KEY_TYPE(t):
    return TectonValidationError(
        f"Invalid type of join keys '{t}'. Keys must be an instance of [pyspark.sql.dataframe.DataFrame, pandas.DataFrame]."
    )


def DUPLICATED_COLS_IN_KEYS(t):
    return TectonValidationError(f"Argument keys {t} have duplicated column names. ")


ATHENA_COMPUTE_ONLY_SUPPORTED_IN_LIVE_WORKSPACE = TectonValidationError(
    "Athena compute can only be used in live workspaces. Current workspace is not live. Please use a different compute mode or switch to a live workspace."
)

ATHENA_COMPUTE_NOT_SUPPORTED_IN_LOCAL_MODE = TectonValidationError(
    "Athena compute can only be used in on applied Tecton objects in live workspaces. Using a locally defined Tecton object is not currently supported with Athena."
)

ATHENA_COMPUTE_MOCK_SOURCES_UNSUPPORTED = TectonValidationError(
    "Athena compute can only be used with materialized data in live workspaces. Using mock data is not supported with Athena."
)

SNOWFLAKE_COMPUTE_MOCK_SOURCES_UNSUPPORTED = TectonValidationError(
    "Using mock data in `get_historical_features` is not supported with Snowflake."
)


def INVALID_USAGE_FOR_LOCAL_TECTON_OBJECT(function_name: str):
    return TectonValidationError(
        f"`{function_name}` can only be called on Tecton objects that have been applied to a Tecton workspace. This object was defined locally."
    )


def INVALID_USAGE_FOR_REMOTE_TECTON_OBJECT(function_name: str):
    return TectonValidationError(
        f"`{function_name}` can only be called on Tecton objects that have been defined locally. This object was retrieved from a Tecton workspace."
    )


def INVALID_USAGE_FOR_LOCAL_FEATURE_TABLE_OBJECT(function_name: str):
    return TectonValidationError(
        f"`{function_name}` can only be called on Feature Tables that have been applied to a Tecton workspace. This object was defined locally, which means this Feature Table cannot be materialized. Feature Tables require materialization in order to ingest features and perform feature retrieval."
    )


def CANNOT_USE_LOCAL_RUN_ON_REMOTE_OBJECT(function_name: str):
    return TectonValidationError(
        f"`{function_name}` can only be called on locally defined Tecton objects. This object was retrieved from a Tecton workpace. Please use `run()` instead."
    )


def INVALID_NUMBER_OF_FEATURE_VIEW_INPUTS(num_sources: int, num_inputs: int):
    return TectonValidationError(
        f"Number of Feature View Inputs ({num_inputs}) should match the number of Data Sources ({num_sources}) in the definition."
    )


def SCHEMAS_DO_NOT_MATCH(schema: schema_pb2.Schema, derived_schema: schema_pb2.Schema):
    return TectonValidationError(
        f"The provided schema does not match the derived schema.\nProvided schema: {Schema(schema)}\nDerived schema: {Schema(derived_schema)}"
    )


def INGESTAPI_USER_ERROR(status_code: int, reason: str, error_message: str):
    if (
        isinstance(error_message, dict)
        and "requestError" in error_message
        and "errorMessage" in error_message["requestError"]
    ):
        error_message = error_message["requestError"]["errorMessage"]
    return TectonValidationError(
        f"Received {status_code} {reason} from Stream IngestAPI. Error Message: \n {error_message}"
    )


def AGGREGATION_INTERVAL_SET_COMPACTION():
    return TectonValidationError("Feature views with stream compaction enabled cannot have `aggregation_interval` set.")


def COMPACTION_TIME_WINDOW_SERIES_UNSUPPORTED():
    return TectonValidationError(
        "Aggregations using a TimeWindowSeries window are not supported when `compaction_enabled=True`."
    )


def TOO_MANY_TRANSFORMATION_CONTEXTS(func_name: str):
    return TectonValidationError(f"Only 1 Context Parameter can be passed into transformation {func_name}.")


TRANSFORMATION_CONTEXT_NOT_SUPPORTED = TectonValidationError(
    "Transformation Contexts are only supported on Batch, Stream and Realtime Feature Views."
)


def TRANSFORMATION_CONTEXT_NAME_NOT_FOUND(context_parameter_name: str, user_function: str):
    return TectonValidationError(
        f'Could not find context parameter "{context_parameter_name}" for function {user_function}.'
    )


# New Read API
GET_FEATURES_FOR_EVENTS_UNSUPPORTED = TectonValidationError(
    "get_features_for_events() is only supported for SPARK or RIFT Compute Modes."
)

GET_FEATURES_IN_RANGE_UNSUPPORTED = TectonValidationError(
    "get_features_in_range() is only supported for SPARK or RIFT Compute Modes."
)

GET_PARTIAL_AGGREGATES_UNSUPPORTED_NON_AGGREGATE = TectonValidationError(
    "get_partial_aggregates() is only supported for Feature Views with Tecton Managed Aggregations."
)

GET_PARTIAL_AGGREGATES_UNSUPPORTED_COMPACTED = TectonValidationError(
    "get_partial_aggregates() is only supported for Aggregate Feature Views with Compaction Disabled."
)

GET_PARTIAL_AGGREGATES_UNSUPPORTED_CONTINUOUS = TectonValidationError(
    'Tecton does not create partial aggregates for Feature Views with "Continuous" Stream Processing Mode.'
)

RUN_TRANSFORMATION_UNSUPPORTED = TectonValidationError(
    "run_transformation() for Batch and Stream Feature Views is only supported for SPARK and RIFT Compute Modes."
)


def READ_API_DEPRECATION_TEMPLATE(old_method, new_method):
    return f"{old_method}() has been deprecated in Tecton 0.9 and will be fully removed in the next major SDK release. Please use {new_method}() instead. See https://docs.tecton.ai/docs/reading-feature-data/reading-feature-data-for-training/offline-retrieval-methods for more information."


GET_HISTORICAL_FEATURES_DEPRECATED_SPINE = READ_API_DEPRECATION_TEMPLATE(
    "get_historical_features", "get_features_for_events"
)

GET_HISTORICAL_FEATURES_DEPRECATED_TIME_RANGE = READ_API_DEPRECATION_TEMPLATE(
    "get_historical_features", "get_features_in_range"
)

RUN_DEPRECATED_TRANSFORMATION = READ_API_DEPRECATION_TEMPLATE("run", "run_transformation")

RUN_DEPRECATED_PARTIAL_AGGS = READ_API_DEPRECATION_TEMPLATE("run", "get_partial_aggregates")

RUN_DEPRECATED_FULL_AGG = READ_API_DEPRECATION_TEMPLATE("run", "get_features_in_range")

GET_SPINE_DF_DEPRECATED = READ_API_DEPRECATION_TEMPLATE("get_spine_dataframe", "get_events_dataframe")


class FeaturesRequired(TectonValidationError):
    def __init__(self, feature_view_name: str):
        super().__init__(f"The 'features' parameter for {feature_view_name} must be set.")


BUILD_ARGS_INTERNAL_ERROR = TectonInternalError(
    "_build_args() is for internal use only and can only be called on local objects"
)

UNSUPPORTED_FRAMEWORK_VERSION = RuntimeError(
    "The existing feature definitions have been applied with an older SDK. Please downgrade the Tecton SDK or upgrade the feature definitions."
)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Python package for random data generation.
"""

import sys
from functools import wraps

from pyspark.mllib.common import callMLlibFunc


__all__ = ['RandomRDDs', ]


def toArray(f):
    @wraps(f)
    def func(sc, *a, **kw):
        rdd = f(sc, *a, **kw)
        return rdd.map(lambda vec: vec.toArray())
    return func


class RandomRDDs(object):
    """
    Generator methods for creating RDDs comprised of i.i.d samples from
    some distribution.

    .. versionadded:: 1.1.0
    """

    @staticmethod
    def uniformRDD(sc, size, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of i.i.d. samples from the
        uniform distribution U(0.0, 1.0).

        To transform the distribution in the generated RDD from U(0.0, 1.0)
        to U(a, b), use
        ``RandomRDDs.uniformRDD(sc, n, p, seed).map(lambda v: a + (b - a) * v)``

        .. versionadded:: 1.1.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            used to create the RDD.
        size : int
            Size of the RDD.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of float comprised of i.i.d. samples ~ `U(0.0, 1.0)`.

        Examples
        --------
        >>> x = RandomRDDs.uniformRDD(sc, 100).collect()
        >>> len(x)
        100
        >>> max(x) <= 1.0 and min(x) >= 0.0
        True
        >>> RandomRDDs.uniformRDD(sc, 100, 4).getNumPartitions()
        4
        >>> parts = RandomRDDs.uniformRDD(sc, 100, seed=4).getNumPartitions()
        >>> parts == sc.defaultParallelism
        True
        """
        return callMLlibFunc("uniformRDD", sc._jsc, size, numPartitions, seed)

    @staticmethod
    def normalRDD(sc, size, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of i.i.d. samples from the standard normal
        distribution.

        To transform the distribution in the generated RDD from standard normal
        to some other normal N(mean, sigma^2), use
        ``RandomRDDs.normal(sc, n, p, seed).map(lambda v: mean + sigma * v)``

        .. versionadded:: 1.1.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            used to create the RDD.
        size : int
            Size of the RDD.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of float comprised of i.i.d. samples ~ N(0.0, 1.0).

        Examples
        --------
        >>> x = RandomRDDs.normalRDD(sc, 1000, seed=1)
        >>> stats = x.stats()
        >>> stats.count()
        1000
        >>> abs(stats.mean() - 0.0) < 0.1
        True
        >>> abs(stats.stdev() - 1.0) < 0.1
        True
        """
        return callMLlibFunc("normalRDD", sc._jsc, size, numPartitions, seed)

    @staticmethod
    def logNormalRDD(sc, mean, std, size, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of i.i.d. samples from the log normal
        distribution with the input mean and standard distribution.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            used to create the RDD.
        mean : float
            mean for the log Normal distribution
        std : float
            std for the log Normal distribution
        size : int
            Size of the RDD.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        RDD of float comprised of i.i.d. samples ~ log N(mean, std).

        Examples
        --------
        >>> from math import sqrt, exp
        >>> mean = 0.0
        >>> std = 1.0
        >>> expMean = exp(mean + 0.5 * std * std)
        >>> expStd = sqrt((exp(std * std) - 1.0) * exp(2.0 * mean + std * std))
        >>> x = RandomRDDs.logNormalRDD(sc, mean, std, 1000, seed=2)
        >>> stats = x.stats()
        >>> stats.count()
        1000
        >>> abs(stats.mean() - expMean) < 0.5
        True
        >>> from math import sqrt
        >>> abs(stats.stdev() - expStd) < 0.5
        True
        """
        return callMLlibFunc("logNormalRDD", sc._jsc, float(mean), float(std),
                             size, numPartitions, seed)

    @staticmethod
    def poissonRDD(sc, mean, size, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of i.i.d. samples from the Poisson
        distribution with the input mean.

        .. versionadded:: 1.1.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        mean : float
            Mean, or lambda, for the Poisson distribution.
        size : int
            Size of the RDD.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of float comprised of i.i.d. samples ~ Pois(mean).

        Examples
        --------
        >>> mean = 100.0
        >>> x = RandomRDDs.poissonRDD(sc, mean, 1000, seed=2)
        >>> stats = x.stats()
        >>> stats.count()
        1000
        >>> abs(stats.mean() - mean) < 0.5
        True
        >>> from math import sqrt
        >>> abs(stats.stdev() - sqrt(mean)) < 0.5
        True
        """
        return callMLlibFunc("poissonRDD", sc._jsc, float(mean), size, numPartitions, seed)

    @staticmethod
    def exponentialRDD(sc, mean, size, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of i.i.d. samples from the Exponential
        distribution with the input mean.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        mean : float
            Mean, or 1 / lambda, for the Exponential distribution.
        size : int
            Size of the RDD.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of float comprised of i.i.d. samples ~ Exp(mean).

        Examples
        --------
        >>> mean = 2.0
        >>> x = RandomRDDs.exponentialRDD(sc, mean, 1000, seed=2)
        >>> stats = x.stats()
        >>> stats.count()
        1000
        >>> abs(stats.mean() - mean) < 0.5
        True
        >>> from math import sqrt
        >>> abs(stats.stdev() - sqrt(mean)) < 0.5
        True
        """
        return callMLlibFunc("exponentialRDD", sc._jsc, float(mean), size, numPartitions, seed)

    @staticmethod
    def gammaRDD(sc, shape, scale, size, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of i.i.d. samples from the Gamma
        distribution with the input shape and scale.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        shape : float
            shape (> 0) parameter for the Gamma distribution
        scale : float
            scale (> 0) parameter for the Gamma distribution
        size : int
            Size of the RDD.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of float comprised of i.i.d. samples ~ Gamma(shape, scale).

        Examples
        --------
        >>> from math import sqrt
        >>> shape = 1.0
        >>> scale = 2.0
        >>> expMean = shape * scale
        >>> expStd = sqrt(shape * scale * scale)
        >>> x = RandomRDDs.gammaRDD(sc, shape, scale, 1000, seed=2)
        >>> stats = x.stats()
        >>> stats.count()
        1000
        >>> abs(stats.mean() - expMean) < 0.5
        True
        >>> abs(stats.stdev() - expStd) < 0.5
        True
        """
        return callMLlibFunc("gammaRDD", sc._jsc, float(shape),
                             float(scale), size, numPartitions, seed)

    @staticmethod
    @toArray
    def uniformVectorRDD(sc, numRows, numCols, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of vectors containing i.i.d. samples drawn
        from the uniform distribution U(0.0, 1.0).

        .. versionadded:: 1.1.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        numRows : int
            Number of Vectors in the RDD.
        numCols : int
            Number of elements in each Vector.
        numPartitions : int, optional
            Number of partitions in the RDD.
        seed : int, optional
            Seed for the RNG that generates the seed for the generator in each partition.

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of Vector with vectors containing i.i.d samples ~ `U(0.0, 1.0)`.

        Examples
        --------
        >>> import numpy as np
        >>> mat = np.matrix(RandomRDDs.uniformVectorRDD(sc, 10, 10).collect())
        >>> mat.shape
        (10, 10)
        >>> mat.max() <= 1.0 and mat.min() >= 0.0
        True
        >>> RandomRDDs.uniformVectorRDD(sc, 10, 10, 4).getNumPartitions()
        4
        """
        return callMLlibFunc("uniformVectorRDD", sc._jsc, numRows, numCols, numPartitions, seed)

    @staticmethod
    @toArray
    def normalVectorRDD(sc, numRows, numCols, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of vectors containing i.i.d. samples drawn
        from the standard normal distribution.

        .. versionadded:: 1.1.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        numRows : int
            Number of Vectors in the RDD.
        numCols : int
            Number of elements in each Vector.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of Vector with vectors containing i.i.d. samples ~ `N(0.0, 1.0)`.

        Examples
        --------
        >>> import numpy as np
        >>> mat = np.matrix(RandomRDDs.normalVectorRDD(sc, 100, 100, seed=1).collect())
        >>> mat.shape
        (100, 100)
        >>> abs(mat.mean() - 0.0) < 0.1
        True
        >>> abs(mat.std() - 1.0) < 0.1
        True
        """
        return callMLlibFunc("normalVectorRDD", sc._jsc, numRows, numCols, numPartitions, seed)

    @staticmethod
    @toArray
    def logNormalVectorRDD(sc, mean, std, numRows, numCols, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of vectors containing i.i.d. samples drawn
        from the log normal distribution.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        mean : float
            Mean of the log normal distribution
        std : float
            Standard Deviation of the log normal distribution
        numRows : int
            Number of Vectors in the RDD.
        numCols : int
            Number of elements in each Vector.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of Vector with vectors containing i.i.d. samples ~ log `N(mean, std)`.

        Examples
        --------
        >>> import numpy as np
        >>> from math import sqrt, exp
        >>> mean = 0.0
        >>> std = 1.0
        >>> expMean = exp(mean + 0.5 * std * std)
        >>> expStd = sqrt((exp(std * std) - 1.0) * exp(2.0 * mean + std * std))
        >>> m = RandomRDDs.logNormalVectorRDD(sc, mean, std, 100, 100, seed=1).collect()
        >>> mat = np.matrix(m)
        >>> mat.shape
        (100, 100)
        >>> abs(mat.mean() - expMean) < 0.1
        True
        >>> abs(mat.std() - expStd) < 0.1
        True
        """
        return callMLlibFunc("logNormalVectorRDD", sc._jsc, float(mean), float(std),
                             numRows, numCols, numPartitions, seed)

    @staticmethod
    @toArray
    def poissonVectorRDD(sc, mean, numRows, numCols, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of vectors containing i.i.d. samples drawn
        from the Poisson distribution with the input mean.

        .. versionadded:: 1.1.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        mean : float
            Mean, or lambda, for the Poisson distribution.
        numRows : float
            Number of Vectors in the RDD.
        numCols : int
            Number of elements in each Vector.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`)
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of Vector with vectors containing i.i.d. samples ~ Pois(mean).

        Examples
        --------
        >>> import numpy as np
        >>> mean = 100.0
        >>> rdd = RandomRDDs.poissonVectorRDD(sc, mean, 100, 100, seed=1)
        >>> mat = np.mat(rdd.collect())
        >>> mat.shape
        (100, 100)
        >>> abs(mat.mean() - mean) < 0.5
        True
        >>> from math import sqrt
        >>> abs(mat.std() - sqrt(mean)) < 0.5
        True
        """
        return callMLlibFunc("poissonVectorRDD", sc._jsc, float(mean), numRows, numCols,
                             numPartitions, seed)

    @staticmethod
    @toArray
    def exponentialVectorRDD(sc, mean, numRows, numCols, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of vectors containing i.i.d. samples drawn
        from the Exponential distribution with the input mean.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        mean : float
            Mean, or 1 / lambda, for the Exponential distribution.
        numRows : int
            Number of Vectors in the RDD.
        numCols : int
            Number of elements in each Vector.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`)
        seed : int, optional
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of Vector with vectors containing i.i.d. samples ~ Exp(mean).

        Examples
        --------
        >>> import numpy as np
        >>> mean = 0.5
        >>> rdd = RandomRDDs.exponentialVectorRDD(sc, mean, 100, 100, seed=1)
        >>> mat = np.mat(rdd.collect())
        >>> mat.shape
        (100, 100)
        >>> abs(mat.mean() - mean) < 0.5
        True
        >>> from math import sqrt
        >>> abs(mat.std() - sqrt(mean)) < 0.5
        True
        """
        return callMLlibFunc("exponentialVectorRDD", sc._jsc, float(mean), numRows, numCols,
                             numPartitions, seed)

    @staticmethod
    @toArray
    def gammaVectorRDD(sc, shape, scale, numRows, numCols, numPartitions=None, seed=None):
        """
        Generates an RDD comprised of vectors containing i.i.d. samples drawn
        from the Gamma distribution.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        sc : :py:class:`pyspark.SparkContext`
            SparkContext used to create the RDD.
        shape : float
            Shape (> 0) of the Gamma distribution
        scale : float
            Scale (> 0) of the Gamma distribution
        numRows : int
            Number of Vectors in the RDD.
        numCols : int
            Number of elements in each Vector.
        numPartitions : int, optional
            Number of partitions in the RDD (default: `sc.defaultParallelism`).
        seed : int, optional,
            Random seed (default: a random long integer).

        Returns
        -------
        :py:class:`pyspark.RDD`
            RDD of Vector with vectors containing i.i.d. samples ~ Gamma(shape, scale).

        Examples
        --------
        >>> import numpy as np
        >>> from math import sqrt
        >>> shape = 1.0
        >>> scale = 2.0
        >>> expMean = shape * scale
        >>> expStd = sqrt(shape * scale * scale)
        >>> mat = np.matrix(RandomRDDs.gammaVectorRDD(sc, shape, scale, 100, 100, seed=1).collect())
        >>> mat.shape
        (100, 100)
        >>> abs(mat.mean() - expMean) < 0.1
        True
        >>> abs(mat.std() - expStd) < 0.1
        True
        """
        return callMLlibFunc("gammaVectorRDD", sc._jsc, float(shape), float(scale),
                             numRows, numCols, numPartitions, seed)


def _test():
    import doctest
    from pyspark.sql import SparkSession
    globs = globals().copy()
    # The small batch size here ensures that we see multiple batches,
    # even in these small test examples:
    spark = SparkSession.builder\
        .master("local[2]")\
        .appName("mllib.random tests")\
        .getOrCreate()
    globs['sc'] = spark.sparkContext
    (failure_count, test_count) = doctest.testmod(globs=globs, optionflags=doctest.ELLIPSIS)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



import typing
from typing import Sequence
from typing import Tuple

from tecton_core import conf


if typing.TYPE_CHECKING:
    import snowflake.snowpark


def register_temp_views_or_tables(views: Sequence[Tuple[str, str]], snowpark_session: "snowflake.snowpark.Session"):
    for view_name, view_sql in views:
        # Using temporary tables provides better performance than temporary views
        temp_object = conf.get_or_none("QUERYTREE_SHORT_SQL_TYPE")
        if not temp_object or temp_object.upper() not in ["VIEW", "TABLE"]:
            msg = f" {temp_object} is not a valid setting for QUERYTREE_SHORT_SQL_TYPE. Must be 'view' or 'table'."
            raise Exception(msg)
        snowpark_session.sql(f"CREATE OR REPLACE TEMPORARY {temp_object.upper()} {view_name} AS ({view_sql})").collect()
